# torchcell/graph/graph.py
# [[torchcell.graph.graph]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/graph/graph.py
# Test file: torchcell/graph/test_graph.py

import glob
import gzip
import json
import logging
import os
import os.path as osp
import pickle
import shutil
import tarfile
import time
from collections import defaultdict
from datetime import datetime
from itertools import product
from typing import Set

import gffutils
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import networkx as nx
import pandas as pd
from attrs import define, field
from Bio import Seq, SeqIO
from Bio.SeqRecord import SeqRecord
from gffutils import FeatureDB
from gffutils.feature import Feature
from matplotlib.patches import Patch
from sortedcontainers import SortedDict, SortedSet
from torch_geometric.data import download_url
from torch_geometric.utils import from_networkx
from tqdm import tqdm

from torchcell.sequence import GeneSet, Genome, ParsedGenome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome
import torchcell
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

# BUG switching to genome for GO, this will create training issues with sql,
# but need genome for created graphs.
@define
class SCerevisiaeGraph:
    data_root: str = field(init=True, repr=False, default="data/sgd/genome")
    # genome: Genome = field(init=True, repr=False, default=None)
    genome: GeneSet = field(init=True, repr=False, default=None)
    json_files: list[str] = field(init=False, repr=False, default=None)
    G_raw: nx.Graph = field(init=False, repr=False, default=None)
    G_gene: nx.Graph = field(init=False, repr=False, default=None)
    G_physical: nx.Graph = field(init=False, repr=False, default=None)
    G_genetic: nx.Graph = field(init=False, repr=False, default=None)
    G_regulatory: nx.DiGraph = field(init=False, repr=False, default=None)
    G_go: nx.DiGraph = field(init=False, repr=False, default=None)
    _all_go_terms = field(init=False, repr=False, default=None)
    _go_to_genes = field(init=False, repr=False, default=None)
    # parsed_genome: ParsedGenome = field(init=True, repr=False, default=None)

    def __attrs_post_init__(self) -> None:
        self.json_files = [f"{gene}.json" for gene in self.genome.gene_set]

        graph_dir = osp.join(self.data_root, "graph")
        if not osp.exists(graph_dir):
            os.makedirs(graph_dir)
        # GO
        go_graph_path = osp.join(graph_dir, "G_go.pkl")
        if osp.exists(go_graph_path):
            self.G_go = self.load_graph("G_go")
        else:
            self.read_raw()
            self.G_go = self.create_G_go()
            # HACK save and unsave
            self.save_graph(self.G_go, "G_go")
        # # Gene Graph
        # gene_graph_path = osp.join(graph_dir, "G_gene.pkl")
        # if osp.exists(gene_graph_path):
        #     self.G_gene = self.load_graph("G_gene")
        # else:
        #     self.G_gene = self.add_gene_protein_overview(
        #         G_raw=self.G_raw, G_gene=nx.Graph()
        #     )
        #     self.save_graph(self.G_gene, "G_gene")

        # Physical Graph
        physical_graph_path = osp.join(graph_dir, "G_physical.pkl")
        if osp.exists(physical_graph_path):
            self.G_physical = self.load_graph("G_physical")
        else:
            self.read_raw()
            self.G_physical = self.add_physical_edges(
                G_raw=self.G_raw, G_physical=nx.Graph()
            )
            self.save_graph(self.G_physical, "G_physical")

        # # Genetic Graph
        # genetic_graph_path = osp.join(graph_dir, "G_genetic.pkl")
        # if osp.exists(genetic_graph_path):
        #     self.G_genetic = self.load_graph("G_genetic")
        # else:
        #     self.G_genetic = self.add_genetic_edges(
        #         G_raw=self.G_raw, G_genetic=nx.Graph()
        #     )
        #     self.save_graph(self.G_genetic, "G_genetic")

        # Regulatory Graph
        regulatory_graph_path = osp.join(graph_dir, "G_regulatory.pkl")
        if osp.exists(regulatory_graph_path):
            self.G_regulatory = self.load_graph("G_regulatory")
        else:
            self.read_raw()
            self.G_regulatory = self.add_regulatory_edges(G_raw=self.G_raw)
            self.save_graph(self.G_regulatory, "G_regulatory")

        # Clean up sql connections
        self.genome = self.parse_genome(self.genome)

    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        data = {}
        data["gene_set"] = genome.gene_set
        return ParsedGenome(**data)

    def read_raw(self):
        # Raw Graph
        # HACK Loading the raw graph is what takes a lot of time, so this should only
        # be called if the other small graphs don't exist
        # keeping for now for quick dev
        graph_dir = osp.join(self.data_root, "graph")
        if not osp.exists(graph_dir):
            os.makedirs(graph_dir)
        raw_graph_path = osp.join(graph_dir, "G_raw.pkl")
        if osp.exists(raw_graph_path):
            self.G_raw = self.load_graph("G_raw")
        else:
            self.G_raw = self.add_json_data_to_graph(
                data_root=self.data_root, json_files=self.json_files
            )
            self.save_graph(self.G_raw, "G_raw")

    def save_graph(self, graph, graph_name):
        dir = osp.join(self.data_root, "graph")
        if not osp.exists(dir):
            os.makedirs(dir)
        path = osp.join(dir, f"{graph_name}.pkl")

        with open(path, "wb") as f:
            pickle.dump(graph, f)

    def load_graph(self, graph_name):
        dir = osp.join(self.data_root, "graph")
        path = osp.join(dir, f"{graph_name}.pkl")

        if not os.path.exists(path):
            print(f"Graph file {path} not found!")
            return None

        with open(path, "rb") as f:
            graph = pickle.load(f)
        return graph

    @staticmethod
    def add_json_data_to_graph(data_root: str, json_files: list[str]) -> nx.Graph:
        G = nx.Graph()  # This is the node graph

        for i, json_file in tqdm(enumerate(json_files)):
            with open(osp.join(data_root, "genes", json_file)) as file:
                data = json.load(file)
                node_name = json_file.split(".")[0]
                G.add_node(node_name, **data)
                if i > 1000:
                    # break
                    pass

        return G

    # GO
    @property
    def all_go_terms(self) -> GeneSet:
        """Collects a SortedSet of all GO terms for all genes in genome.gene_set."""
        if self._all_go_terms is None:
            all_go_terms = GeneSet()
            missing_go_terms = set()
            go_dag = self.genome.go_dag
            for gene in self.genome.gene_set:
                go_details = self.G_raw.nodes[gene].get("go_details", [])
                for detail in go_details:
                    go_id = detail["go"]["go_id"]
                    if go_id in go_dag and not go_dag[go_id].is_obsolete:
                        all_go_terms.add(go_id)
                    else:
                        log.warning(
                            f"GO term {go_id} not found in go_dag for gene {gene}. "
                            "Most likely deprecated."
                        )
                        print(f"deprecated: {go_id}")
                        missing_go_terms.add(go_id)
            log.warning(f"Missing GO terms: {missing_go_terms}")
            self._all_go_terms = all_go_terms
            return self._all_go_terms
        else:  # already computed
            return self._all_go_terms

    @property
    def go_to_genes(self) -> SortedDict[str, GeneSet]:
        if self._go_to_genes is None:
            go_to_genes_dict = SortedDict()

            # Iterate through all genes in genome.gene_set
            for gene in self.genome.gene_set:
                go_details = self.G_raw.nodes[gene].get("go_details", [])

                # Iterate through all GO details for the current gene
                for detail in go_details:
                    go_id = detail["go"]["go_id"]
                    if go_id not in go_to_genes_dict:
                        go_to_genes_dict[go_id] = GeneSet()
                    go_to_genes_dict[go_id].add(gene)

            self._go_to_genes = go_to_genes_dict
            return go_to_genes_dict
        else:
            return self._go_to_genes

    def create_go_subgraph(self, go_terms, go_dag):
        G = nx.DiGraph()  # Using a directed graph for GO hierarchy

        for go_id in go_terms:
            node_data = {
                "id": go_dag[go_id].id,
                "item_id": go_dag[go_id].item_id,
                "name": go_dag[go_id].name,
                "namespace": go_dag[go_id].namespace,
                "level": go_dag[go_id].level,
                "depth": go_dag[go_id].depth,
                "is_obsolete": go_dag[go_id].is_obsolete,
                "alt_ids": go_dag[go_id].alt_ids,
                "gene_set": self.go_to_genes.get(go_id, None),
                "genes": {},
            }
            # Add data from raw
            for gene in node_data["gene_set"]:
                for go_details in self.G_raw.nodes[gene]["go_details"]:
                    if go_details["go"]["go_id"] == go_id:
                        node_data["genes"][gene] = {}
                        node_data["genes"][gene]["go_details"] = go_details
            G.add_node(go_id, **node_data)

            # Add edges
            for parent in go_dag[go_id].parents:
                G.add_edge(go_id, parent.id)

        # Remove nodes that have no data. These are redundant nodes.
        nodes_to_remove = [
            node
            for node, data in G.nodes(data=True)
            if all(value is None for value in data.values())
        ]
        for node in nodes_to_remove:
            G.remove_node(node)

        # Assert there's only one node at level 0
        num_level_0_nodes = len(
            [node for node in G.nodes(data=True) if node[1]["level"] == 0]
        )
        assert (
            num_level_0_nodes == 1
        ), "There should be only one root node for a GO subgraph"
        root_node = [node for node, data in G.nodes(data=True) if data["level"] == 0][0]

        # Check for each level starting from level 0, if all nodes have a path to root
        # If not, make a direct edge from the node to root
        max_level = max(data["level"] for _, data in G.nodes(data=True))
        for level in range(0, max_level + 1):
            nodes_at_level = [
                node for node, data in G.nodes(data=True) if data["level"] == level
            ]
            for node in nodes_at_level:
                if not nx.has_path(G, node, root_node):
                    G.add_edge(node, root_node)

        return G

    @staticmethod
    def combine_with_super_node(graphs: list[nx.Graph] = None) -> nx.Graph:
        G_combined = nx.DiGraph()
        super_node = "GO:ROOT"  # A fictitious super node
        G_combined.add_node(
            super_node, name="GO Super Node", namespace="super_root", level=-1
        )

        for G in graphs:
            # Merge nodes and edges from G into G_combined
            G_combined = nx.compose(G_combined, G)

            # Identify the level 0 nodes of G using the "level" attribute
            level_0_nodes = [
                node for node, data in G.nodes(data=True) if data.get("level") == 0
            ]

            # Link the identified level 0 nodes to the super node
            for node in level_0_nodes:
                G_combined.add_edge(node, super_node)

        return G_combined

    def create_G_go(self) -> nx.DiGraph:
        bp_terms = [
            go_id
            for go_id, term in self.genome.go_dag.items()
            if term.namespace == "biological_process" and go_id in self.all_go_terms
        ]
        mf_terms = [
            go_id
            for go_id, term in self.genome.go_dag.items()
            if term.namespace == "molecular_function" and go_id in self.all_go_terms
        ]
        cc_terms = [
            go_id
            for go_id, term in self.genome.go_dag.items()
            if term.namespace == "cellular_component" and go_id in self.all_go_terms
        ]

        # construct individual subgraphs
        bp_subgraph = self.create_go_subgraph(bp_terms, self.genome.go_dag)
        mf_subgraph = self.create_go_subgraph(mf_terms, self.genome.go_dag)
        cc_subgraph = self.create_go_subgraph(cc_terms, self.genome.go_dag)

        # construct whole GO with super node
        G_go = self.combine_with_super_node([bp_subgraph, mf_subgraph, cc_subgraph])
        return G_go

    # Edge Features
    @staticmethod
    def add_physical_edges(G_raw: nx.Graph, G_physical: nx.Graph) -> nx.Graph:
        for node_name, node_data in G_raw.nodes(data=True):
            if "interaction_details" in node_data and isinstance(
                node_data["interaction_details"], list
            ):
                for interaction in node_data["interaction_details"]:
                    if interaction.get("interaction_type") == "Physical":
                        node1 = interaction["locus1"]["format_name"]
                        node2 = interaction["locus2"]["format_name"]
                        if node1 in G_raw and node2 in G_raw:
                            # TODO Need a mechanism to process the interaction data.
                            # Commented out to work with cell data.
                            G_physical.add_edge(node1, node2, **interaction)
                            G_physical.add_edge(node1, node2)
        return G_physical

    @staticmethod
    def add_genetic_edges(G_raw: nx.Graph, G_genetic: nx.Graph) -> nx.Graph:
        for node_name, node_data in G_raw.nodes(data=True):
            if "interaction_details" in node_data and isinstance(
                node_data["interaction_details"], list
            ):
                for interaction in node_data["interaction_details"]:
                    if interaction.get("interaction_type") == "Genetic":
                        node1 = interaction["locus1"]["format_name"]
                        node2 = interaction["locus2"]["format_name"]
                        if node1 in G_raw and node2 in G_raw:
                            G_genetic.add_edge(node1, node2, **interaction)
        return G_genetic

    @staticmethod
    def add_regulatory_edges(
        G_raw: nx.Graph, G_regulatory: nx.DiGraph = None
    ) -> nx.DiGraph:
        """
        Add regulatory edges from the raw graph to the regulatory graph.
        """
        if G_regulatory is None:
            G_regulatory = nx.DiGraph()

        for node_name, node_data in G_raw.nodes(data=True):
            if "regulation_details" in node_data and isinstance(
                node_data["regulation_details"], list
            ):
                for regulation in node_data["regulation_details"]:
                    # Extract details
                    locus1_name = regulation["locus1"]["format_name"]
                    locus2_name = regulation["locus2"]["format_name"]

                    # Add nodes
                    G_regulatory.add_node(locus1_name, **regulation["locus1"])
                    G_regulatory.add_node(locus2_name, **regulation["locus2"])

                    # Add directed edge
                    G_regulatory.add_edge(locus1_name, locus2_name, **regulation)

        return G_regulatory

    # Node Features
    @staticmethod
    def add_gene_protein_overview(G_raw: nx.Graph, G_gene: nx.Graph) -> nx.Graph:
        for node_name, node_data in G_raw.nodes(data=True):
            protein_overview_template = {
                "length": None,
                "molecular_weight": None,
                "pi": None,
                "median_value": None,
                "median_abs_dev_value": None,
            }
            if "protein_overview" in node_data["locus"]:
                protein_overview = protein_overview_template.copy()
                protein_overview["length"] = node_data["locus"]["protein_overview"].get(
                    "length"
                )
                protein_overview["molecular_weight"] = node_data["locus"][
                    "protein_overview"
                ].get("molecular_weight")
                protein_overview["pi"] = node_data["locus"]["protein_overview"].get(
                    "pi"
                )
                protein_overview["median_value"] = node_data["locus"][
                    "protein_overview"
                ].get("median_value")
                protein_overview["median_abs_dev_value"] = node_data["locus"][
                    "protein_overview"
                ].get("median_abs_dev_value")
                G_gene.add_node(node_name, **protein_overview)
            else:
                G_gene.add_node(node_name, **protein_overview_template)
        return G_gene


def plot_go_graph(G):
    # Define a color map for namespaces.
    namespace_colors = {
        "super_root": "#122A4B",
        "biological_process": "#FF552E",
        "molecular_function": "#8286C2",
        "cellular_component": "#D23943",
    }

    # Identify nodes without 'namespace'
    nodes_missing_namespace = [
        node for node, attrs in G.nodes(data=True) if "namespace" not in attrs
    ]
    print(f"num nodes_missing_namespace: {len(nodes_missing_namespace)}")

    # Generate the color list for nodes based on their namespace
    node_colors = [
        namespace_colors[G.nodes[node].get("namespace", "missing")]
        for node in G.nodes()
    ]

    max_level = max([data["level"] for node, data in G.nodes(data=True)])
    min_level = min([data["level"] for node, data in G.nodes(data=True)])

    # Calculate node positions based on levels
    level_spacing = 10.0
    horizontal_spacing = 20.0

    new_pos = {}
    for node, data in G.nodes(data=True):
        level = data.get("level")
        nodes_in_level = [n for n, d in G.nodes(data=True) if d.get("level") == level]

        # Calculate horizontal position
        idx_in_level = nodes_in_level.index(node)
        total_nodes_in_level = len(nodes_in_level)
        new_x = (idx_in_level - total_nodes_in_level / 2) * horizontal_spacing

        # Adjust vertical position
        new_y = (max_level - level) * level_spacing - min_level * level_spacing

        new_pos[node] = (new_x, new_y)

    # Plotting
    plt.figure(figsize=(20, 10))
    nx.draw(
        G,
        new_pos,
        with_labels=False,
        node_size=20,
        node_color=node_colors,
        alpha=0.6,
        linewidths=0.5,
        width=0.25,
        edge_color="lightgray",
        arrowsize=5,
    )

    # Add legend
    legend_elements = [
        Patch(
            facecolor=namespace_colors[key],
            edgecolor="black",
            label=key.replace("_", " ").title(),
        )
        for key in namespace_colors
    ]
    plt.legend(
        handles=legend_elements,
        loc="upper right",
        title="Namespaces",
        fontsize=28,
        title_fontsize=28,
    )

    # Annotate levels on the left side
    unique_levels = sorted(list({data["level"] for node, data in G.nodes(data=True)}))

    # Determine the minimum x-coordinate and set the margin
    min_x_coord = min([x for x, y in new_pos.values()])
    margin = 120.0 * horizontal_spacing  # Adjust the multiplier as needed

    for level in unique_levels:
        y_pos = (max_level - level) * level_spacing - min_level * level_spacing
        plt.text(
            x=min_x_coord - margin,  # Use the adjusted x-coordinate
            y=y_pos,
            s=f"level: {level}",
            ha="left",
            va="center",
            fontsize=28,
        )

    plt.title("GO DAG with Forwarded Isolated Nodes and Super Node", fontsize=28)
    plt.tight_layout()
    plt.savefig(
        "./notes/assets/images/dcell-gene-ontology-dag-no-isoloated-with-super-node.png",
        dpi=300,
    )
    plt.close()
    # plt.show()


# TODO add a function that will remove genes not in dmf costanzo and smf costanzo


def filter_go_IGI(G_go: nx.DiGraph) -> nx.DiGraph:
    # Create a copy of the original graph - avoid inplace
    G = G_go.copy()
    nodes_to_remove = []

    for go_term in G.nodes():
        if "gene_set" in G.nodes[go_term] and "genes" in G.nodes[go_term]:
            # Identify genes with "IGI" experiment
            remove_genes = [
                gene
                for gene in G.nodes[go_term]["gene_set"]
                if gene in G.nodes[go_term]["genes"]
                and G.nodes[go_term]["genes"][gene]["go_details"]["experiment"][
                    "display_name"
                ]
                == "IGI"
            ]

            # Remove genes from gene_set and genes attribute
            for gene in remove_genes:
                G.nodes[go_term]["gene_set"].remove(gene)
                del G.nodes[go_term]["genes"][gene]

            # Check if gene_set is now empty
            if len(G.nodes[go_term]["gene_set"]) == 0:
                in_nodes = list(G.predecessors(go_term))
                out_nodes = list(G.successors(go_term))

                # Forward the in edges of go_term to each of its out nodes
                for in_node in in_nodes:
                    for out_node in out_nodes:
                        G.add_edge(in_node, out_node)
                nodes_to_remove.append(go_term)

    # Remove nodes marked for deletion
    log.info(f"IGI nodes removed: {len(nodes_to_remove)}")
    for node in nodes_to_remove:
        G.remove_node(node)

    return G


def filter_by_date(G_go: nx.DiGraph, cutoff_date: str) -> nx.DiGraph:
    # Create a copy of the original graph - avoid inplace
    G = G_go.copy()
    nodes_to_remove = []

    for go_term in G.nodes():
        if "gene_set" in G.nodes[go_term] and "genes" in G.nodes[go_term]:
            # Identify genes whose annotations are after the cutoff date
            remove_genes = [
                gene
                for gene in G.nodes[go_term]["gene_set"]
                if gene in G.nodes[go_term]["genes"]
                and G.nodes[go_term]["genes"][gene]["go_details"]["date_created"]
                > cutoff_date
            ]

            # Remove genes from gene_set and genes attribute
            for gene in remove_genes:
                G.nodes[go_term]["gene_set"].remove(gene)
                del G.nodes[go_term]["genes"][gene]

            # Check if gene_set is now empty
            if len(G.nodes[go_term]["gene_set"]) == 0:
                in_nodes = list(G.predecessors(go_term))
                out_nodes = list(G.successors(go_term))

                # Forward the in edges of go_term to each of its out nodes
                for in_node in in_nodes:
                    for out_node in out_nodes:
                        G.add_edge(in_node, out_node)
                nodes_to_remove.append(go_term)

    # Remove nodes marked for deletion
    log.info(f"Nodes annotated after {cutoff_date} removed: {len(nodes_to_remove)}")
    for node in nodes_to_remove:
        G.remove_node(node)

    return G


def filter_redundant_terms(G_go: nx.DiGraph) -> nx.DiGraph:
    # Create a copy of the original graph - avoid inplace
    G = G_go.copy()

    # Find leaf nodes - nodes with no successors
    leaf_nodes = [node for node in G if len(list(G.successors(node))) == 0]

    # Nodes to check for redundancy
    nodes_to_check = set(leaf_nodes)
    nodes_to_remove = set()

    while nodes_to_check:
        node = nodes_to_check.pop()
        parent_nodes = list(G.predecessors(node))

        for parent in parent_nodes:
            if (
                "gene_set" in G.nodes[parent]
                and "gene_set" in G.nodes[node]
                and set(G.nodes[parent]["gene_set"]) == set(G.nodes[node]["gene_set"])
            ):
                # Add the redundant node to the list for removal
                nodes_to_remove.add(node)

                # Forward the in edges of node to each of its out nodes
                in_nodes = list(G.predecessors(node))
                out_nodes = list(G.successors(node))
                for in_node in in_nodes:
                    for out_node in out_nodes:
                        G.add_edge(in_node, out_node)

            # Add the parent to the nodes to check, only if it's not already marked for removal
            if parent not in nodes_to_remove:
                nodes_to_check.add(parent)

    # Remove nodes marked for deletion
    log.info(f"Redundant nodes removed: {len(nodes_to_remove)}")
    for node in nodes_to_remove:
        if node in G:
            G.remove_node(node)

    return G


def filter_by_contained_genes(G_go: nx.DiGraph, n: int, gene_set: set) -> nx.DiGraph:
    G = G_go.copy()

    def compute_containment(go_term) -> set[str]:
        """Function to compute containment of a given term with its subsequent terms."""

        # Reverse the graph to travel in the opposite direction
        G_reverse = G_go.reverse(copy=False)

        # Find all reachable nodes from the current term in the reversed graph
        reachable_nodes = nx.single_source_shortest_path_length(
            G_reverse, go_term
        ).keys()

        # Get gene sets of all these nodes
        gene_sets = [
            set(G_go.nodes[node].get("gene_set", [])) for node in reachable_nodes
        ]

        # Compute the union of all gene sets
        contained_genes = set.union(*gene_sets)
        if gene_set is not None:
            contained_genes = contained_genes.intersection(gene_set)
        return contained_genes

    # Construct the containment_dictionary
    containment_dictionary = {node: compute_containment(node) for node in G.nodes()}

    # Mark nodes for removal based on containment criteria
    nodes_to_remove = [
        node for node, genes in containment_dictionary.items() if len(genes) < n
    ]

    # Forward the connections of nodes marked for removal
    for node in nodes_to_remove:
        in_nodes = list(G.predecessors(node))
        out_nodes = list(G.successors(node))
        for in_node in in_nodes:
            for out_node in out_nodes:
                if in_node != out_node and not G.has_edge(in_node, out_node):
                    G.add_edge(in_node, out_node)

    # Remove the marked nodes
    log.info(f"Nodes with < {n} contained genes removed: {len(nodes_to_remove)}")
    for node in nodes_to_remove:
        G.remove_node(node)

    return G


def plot_annotation_dates_by_month(G_go: nx.DiGraph):
    # Function to convert date strings to datetime objects
    def convert_to_datetime(date_str):
        return datetime.strptime(date_str, "%Y-%m-%d")

    # Collect all the annotation dates and group by month and year
    dates = defaultdict(int)
    for go_term in G_go.nodes():
        if "genes" in G_go.nodes[go_term]:
            for gene, details in G_go.nodes[go_term]["genes"].items():
                date_obj = convert_to_datetime(details["go_details"]["date_created"])
                month_year = date_obj.strftime("%Y-%m")  # Format as "YYYY-MM"
                dates[month_year] += 1

    # Sort the dates chronologically
    sorted_dates = sorted(dates.items(), key=lambda x: datetime.strptime(x[0], "%Y-%m"))

    # Extract x and y values
    x_vals, y_vals = zip(*sorted_dates)

    # Extract unique years and their first month occurrence for labeling purposes
    year_ticks = [date for idx, date in enumerate(x_vals) if date.endswith("-01")]

    # Plot the histogram
    plt.bar(x_vals, y_vals, edgecolor="k", alpha=0.7)
    plt.title("Gene Ontology Annotation Dates (Grouped by Month)", fontsize=24)
    plt.xlabel("Date (Year)", labelpad=10, fontsize=24)
    plt.ylabel("Number of Annotations", labelpad=10, fontsize=24)

    # Display only the labels for the years using their first month
    plt.xticks(year_ticks, rotation=45, fontsize=16)
    plt.yticks(fontsize=16)

    # Make year ticks longer for distinction
    ax = plt.gca()
    for tick in ax.xaxis.get_major_ticks():
        if tick.label1.get_text() in year_ticks:
            tick.tick1line.set_markersize(10)
            tick.label1.set_fontweight("bold")
    plt.tight_layout()
    plt.savefig(
        "./notes/assets/images/gene-ontology-annotation-dates-grouped-by-month.png",
        dpi=300,
    )
    # plt.show()
    plt.close()


def plot_histogram_of_gene_counts(G_go: nx.DiGraph):
    # Extracting the gene sets for each GO term
    gene_sets = [G_go.nodes[node].get("gene_set", []) for node in G_go.nodes()]

    # Getting the counts of genes for each GO term
    gene_counts = [len(gene_set) for gene_set in gene_sets]

    # Plotting the histogram
    plt.hist(gene_counts, bins=100, alpha=0.7)
    plt.title("Histogram of Gene Counts per GO Term")
    plt.xlabel("Number of Genes")
    plt.ylabel("Number of GO Terms")
    plt.grid(axis="y", linestyle="--")
    plt.tight_layout()
    plt.show()


def plot_histogram_of_contained_gene_counts(
    G_go: nx.DiGraph, gene_set: set = None, show_min_max: bool = False
):
    def compute_containment(go_term):
        """Function to compute containment of a given term with its subsequent terms."""

        # Reverse the graph to travel in the opposite direction
        G_reverse = G_go.reverse(copy=False)

        # Find all reachable nodes from the current term in the reversed graph
        reachable_nodes = nx.single_source_shortest_path_length(
            G_reverse, go_term
        ).keys()

        # Get gene sets of all these nodes
        gene_sets = [
            set(G_go.nodes[node].get("gene_set", [])) for node in reachable_nodes
        ]

        # Compute the union of all gene sets
        contained_genes = set.union(*gene_sets)
        if gene_set is not None:
            contained_genes = contained_genes.intersection(gene_set)
        return contained_genes

    # Construct containment_dictionary
    containment_dictionary = {node: compute_containment(node) for node in G_go.nodes()}

    # Get counts for histogram
    contained_gene_counts = [len(genes) for genes in containment_dictionary.values()]

    # Plotting the histogram
    plt.hist(contained_gene_counts, bins=100, alpha=0.7)
    plt.title("Histogram of Contained Gene Counts per GO Term")
    plt.xlabel("Number of Contained Genes")
    plt.ylabel("Number of GO Terms")

    # Optional vertical dashed lines for min and max values
    if show_min_max:
        plt.axvline(min(contained_gene_counts), color="red", linestyle="--", alpha=0.7)
        plt.axvline(max(contained_gene_counts), color="red", linestyle="--", alpha=0.7)

    plt.grid(axis="y", linestyle="--")
    plt.tight_layout()
    plt.show()


def compare_go_terms(G_raw, G):
    matching_nodes = 0
    total_nodes = len(G_raw.nodes())
    total_new_terms = (
        0  # This will accumulate the total new terms added to G_raw from G
    )

    for node in G_raw.nodes():
        # Extract GO terms from the complex graph G_raw
        go_details_G_raw = G_raw.nodes[node]["go_details"]
        go_terms_G_raw = {
            detail["go"]["go_id"] for detail in go_details_G_raw if "go" in detail
        }

        # Check if the node exists in the simpler graph G
        if node in G.nodes():
            # Extract GO terms from the simple graph G
            go_terms_G = {rec["GO_ID"] for rec in G.nodes[node]["annotations"]}

            # Check if the terms in G are exactly the same as in G_raw
            if go_terms_G == go_terms_G_raw:
                matching_nodes += 1
            else:
                # Calculate the new terms from G that aren't in G_raw
                new_terms = go_terms_G - go_terms_G_raw
                total_new_terms += len(new_terms)
                # Optionally, report the node and the number of new terms
                print(
                    f"Node {node}: {len(new_terms)} new terms from G would be added to G_raw."
                )
        else:
            # Handle the case where a node in G_raw does not exist in G
            print(f"Node {node} from G_raw does not exist in G.")

    print(f"{matching_nodes} out of {total_nodes} nodes have matching GO term sets.")
    print(f"A total of {total_new_terms} new GO terms would be added to G_raw from G.")


def go_gaf_investigation():
    import os
    import random

    import matplotlib.pyplot as plt
    import pandas as pd
    from Bio.UniProt.GOA import gafiterator
    from dotenv import load_dotenv

    from torchcell.datasets.scerevisiae import (
        DmfCostanzo2016Dataset,
        SmfCostanzo2016Dataset,
    )

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), overwrite=True
    )
    genome.drop_chrmt()
    genome.drop_empty_go()
    print(f"Gene Set Length: {len(genome.gene_set)}")
    # Initialize the graph
    G = nx.Graph()

    # Add nodes to the graph
    for gene_name in genome.gene_set:
        G.add_node(gene_name, annotations=[])

    # Define your filters here
    synonym_filter = {"Synonym": genome.gene_set}

    # Function to check if a record matches the given filters
    def record_has(rec, filters):
        for field, values in filters.items():
            if field in rec and not set(rec[field]).isdisjoint(values):
                return True
        return False

    # Read the GAF file and apply filters
    gaf_path = osp.join(DATA_ROOT, "data/sgd.gaf")
    with open(gaf_path) as handle:
        for record in gafiterator(handle):
            # Apply filters
            for synonym in record.get("Synonym", []):
                if synonym in genome.gene_set:
                    G.nodes[synonym]["annotations"].append(record)
                    break  # Once matched, no need to check other synonyms

    # use this len(G.nodes[[node for node in G.nodes][100]]["annotations"])
    graph = SCerevisiaeGraph(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), genome=genome
    )
    graph.G_go
    print()
    compare_go_terms(graph.G_raw, G)


import networkx as nx
import plotly.graph_objects as go


def plotly_go_graph(G):
    # Define color map for namespaces
    namespace_colors = {
        "super_root": "#122A4B",
        "biological_process": "#FF552E",
        "molecular_function": "#8286C2",
        "cellular_component": "#D23943",
        "missing": "#AAAAAA",  # Color for missing namespace
    }

    # Determine levels and layout
    max_level = max(data.get("level", 0) for node, data in G.nodes(data=True))
    min_level = min(data.get("level", 0) for node, data in G.nodes(data=True))
    level_spacing = 2.0  # Increased spacing for levels
    horizontal_spacing = 2.0  # Increased spacing horizontally

    new_pos = {}
    for node, data in G.nodes(data=True):
        level = data.get("level", 0)
        nodes_in_level = [n for n, d in G.nodes(data=True) if d.get("level") == level]
        idx_in_level = nodes_in_level.index(node)
        total_nodes_in_level = len(nodes_in_level)
        new_x = (idx_in_level - total_nodes_in_level / 2) * horizontal_spacing
        new_y = (max_level - level) * level_spacing - min_level * level_spacing
        new_pos[node] = (new_x, new_y)

    # Creating edge traces
    edge_x = []
    edge_y = []
    for edge in G.edges():
        x0, y0 = new_pos[edge[0]]
        x1, y1 = new_pos[edge[1]]
        edge_x.extend([x0, x1, None])
        edge_y.extend([y0, y1, None])

    edge_trace = go.Scatter(
        x=edge_x,
        y=edge_y,
        line=dict(width=0.5, color="#888"),
        hoverinfo="none",
        mode="lines",
    )

    # Creating node traces
    node_x = []
    node_y = []
    node_info = []
    node_color = []
    for node in G.nodes():
        x, y = new_pos[node]
        node_x.append(x)
        node_y.append(y)
        node_data = G.nodes[node]
        name = node_data.get("name", "none")
        level = node_data.get("level", "none")
        namespace = node_data.get("namespace", "missing")
        info_str = f"{node}: name={name}, level={level}"
        node_info.append(info_str)
        node_color.append(namespace_colors.get(namespace, "missing"))

    node_trace = go.Scatter(
        x=node_x,
        y=node_y,
        mode="markers",
        hoverinfo="text",
        text=node_info,
        marker=dict(showscale=False, size=7, color=node_color, line=None),
    )
    # Create the figure
    fig = go.Figure(
        data=[edge_trace, node_trace],
        layout=go.Layout(
            title="Gene Ontology Graph",
            titlefont_size=16,
            showlegend=False,
            hovermode="closest",
            margin=dict(b=20, l=5, r=5, t=40),
            xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            plot_bgcolor="rgba(0,0,0,0)",
        ),
    )

    fig.show()


def main() -> None:
    import os
    import random

    import matplotlib.pyplot as plt
    import pandas as pd
    from dotenv import load_dotenv

    from torchcell.datasets.scerevisiae import (
        DmfCostanzo2016Dataset,
        SmfCostanzo2016Dataset,
    )

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), overwrite=True
    )
    graph = SCerevisiaeGraph(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), genome=genome
    )
    #
    dmf_dataset = DmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e5"),
        preprocess={"duplicate_resolution": "low_dmf_std"},
        # subset_n=100,
    )
    smf_dataset = SmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_smf"),
        preprocess={"duplicate_resolution": "low_std_both"},
        skip_process_file_exist_check=True,
    )
    gene_set = smf_dataset.gene_set.union(dmf_dataset.gene_set)
    #
    print(graph.G_go.number_of_nodes())
    G = graph.G_go.copy()

    ##### Filtering
    # plot_histogram_of_gene_counts(G)
    # plot_histogram_of_contained_gene_counts(G, gene_set=gene_set, show_min_max=True)
    # G = filter_by_date(G, "2018-02-01")
    G = filter_by_date(G, "2017-07-19")
    print(f"After date filter: {G.number_of_nodes()}")
    G = filter_go_IGI(G)
    print(f"After IGI filter: {G.number_of_nodes()}")
    G = filter_redundant_terms(G)
    print(f"After redundant filter: {G.number_of_nodes()}")
    # (filter_by_contained_genes(G, 1, gene_set=gene_set)).number_of_nodes()
    G = filter_by_contained_genes(G, n=6, gene_set=gene_set)
    print(f"After containment filter: {G.number_of_nodes()}")
    # print()
    # plot_go_graph(G)
    print()

    # plot_annotation_dates_by_month(graph.G_go)
    # plot_go_graph(graph.G_go)
    plotly_go_graph(G)


if __name__ == "__main__":
    main()
    # main_go()

    # # TODO maybe remove
    # @staticmethod
    # def parse_genome(genome) -> ParsedGenome:
    #     # BUG we have to do this black magic because when you merge datasets with +
    #     # the genome is None
    #     if genome is None:
    #         return None
    #     else:
    #         data = {}
    #         data["gene_set"] = genome.gene_set
    #         return ParsedGenome(**data)
    #         data["gene_set"] = genome.gene_set
    #         return ParsedGenome(**data)



# torchcell/datasets/scerevisiae/costanzo2016.py
# [[torchcell.datasets.scerevisiae.costanzo2016]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/scerevisiae/costanzo2016.py
# Test file: torchcell/datasets/scerevisiae/test_costanzo2016.py
import json
import logging
import os
import os.path as osp
import pickle
import random
import re
import shutil
import zipfile
from abc import ABC, abstractproperty
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from typing import Literal, Optional

import h5py
import lmdb
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
# import polars as pl
import torch
# from polars import DataFrame, col
from torch_geometric.data import (
    Data,
    DataLoader,
    InMemoryDataset,
    download_url,
    extract_zip,
)
from tqdm import tqdm

from torchcell.data import Dataset
from torchcell.datamodels import ModelStrict
from torchcell.prof import prof, prof_input
from torchcell.sequence import GeneSet

log = logging.getLogger(__name__)


class SmfCostanzo2016Dataset(Dataset):
    url = (
        "https://thecellmap.org/costanzo2016/data_files/"
        "Raw%20genetic%20interaction%20datasets:%20Pair-wise%20interaction%20format.zip"
    )

    def __init__(
        self,
        root: str = "data/scerevisiae/costanzo2016",
        subset_n: int = None,
        preprocess: dict | None = None,
        skip_process_file_exist_check: bool = False,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.subset_n = subset_n
        self._skip_process_file_exist = skip_process_file_exist_check
        self.preprocess = preprocess
        self.preprocess_dir = osp.join(root, "preprocess")
        self._length = None
        self._gene_set = None
        self._df = None
        # Check for existing preprocess config
        existing_config = self.load_preprocess_config()
        if existing_config is not None:
            if existing_config != self.preprocess:
                raise ValueError(
                    "New preprocess does not match existing config."
                    "Delete the processed and process dir for a new Dataset."
                    "Or define a new root."
                )
        super().__init__(root, transform, pre_transform)
        self.env = None

    @property
    def skip_process_file_exist(self):
        return self._skip_process_file_exist

    @property
    def raw_file_names(self) -> list[str]:
        return ["strain_ids_and_single_mutant_fitness.xlsx"]

    @property
    def processed_file_names(self) -> list[str]:
        return "data.lmdb"

    def download(self):
        path = download_url(self.url, self.raw_dir)
        with zipfile.ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(self.raw_dir)
        os.remove(path)

        # Move the contents of the subdirectory to the parent raw directory
        sub_dir = os.path.join(
            self.raw_dir,
            "Data File S1. Raw genetic interaction datasets: Pair-wise interaction format",
        )
        for filename in os.listdir(sub_dir):
            shutil.move(os.path.join(sub_dir, filename), self.raw_dir)
        os.rmdir(sub_dir)

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.processed_dir, "data.lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
        )

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None

    # TODO implement
    @property
    def df(self):
        if osp.exists(osp.join(self.preprocess_dir, "data.csv")):
            self._df = pd.read_csv(osp.join(self.preprocess_dir, "data.csv"))
        return self._df

    @property
    def wt(self):
        wt = {}
        wt["genotype"] = ({"id": None, "intervention": None, "id_full": None},)
        wt["phenotype"] = {
            "observation": {"fitness": 1.0},
            "environment": {"media": "YPD", "temperature": 30},
        }
        data = Data()
        data.genotype = wt["genotype"]
        data.phenotype = wt["phenotype"]
        return data

    def process(self):
        os.makedirs(self.preprocess_dir, exist_ok=True)
        # Process the Excel file for mutant fitness
        excel_file_path = os.path.join(
            self.raw_dir, "strain_ids_and_single_mutant_fitness.xlsx"
        )
        df_excel = pd.read_excel(excel_file_path)

        # Preproecess
        all_data_df = self.preprocess_raw(df_excel, self.preprocess)
        self.save_preprocess_config(self.preprocess)

        # Subset
        if self.subset_n is not None:
            all_data_df = all_data_df.sample(
                n=self.subset_n, random_state=42
            ).reset_index(drop=True)

        # Save preprocssed df - mainly for quick stats
        all_data_df.to_csv(osp.join(self.preprocess_dir, "data.csv"), index=False)

        print("Processing SMF data...")

        # Extract genotypes
        all_data_df["genotype_id"] = all_data_df["Strain ID"].str.split("_").str[0]
        genotypes = all_data_df[["genotype_id", "Strain ID"]].to_dict("records")

        # Create 26° observations
        observations_26 = (
            all_data_df[
                ["Single mutant fitness (26°)", "Single mutant fitness (26°) stddev"]
            ]
            .rename(
                columns={
                    "Single mutant fitness (26°)": "smf",
                    "Single mutant fitness (26°) stddev": "smf_std",
                }
            )
            .to_dict("records")
        )
        environments_26 = [
            {"media": "YPD", "temperature": 26} for _ in range(len(all_data_df))
        ]

        # Create 30° observations
        observations_30 = (
            all_data_df[
                ["Single mutant fitness (30°)", "Single mutant fitness (30°) stddev"]
            ]
            .rename(
                columns={
                    "Single mutant fitness (30°)": "smf",
                    "Single mutant fitness (30°) stddev": "smf_std",
                }
            )
            .to_dict("records")
        )
        environments_30 = [
            {"media": "YPD", "temperature": 30} for _ in range(len(all_data_df))
        ]

        data_list = []
        for genotype, obs_26, env_26, obs_30, env_30 in zip(
            genotypes,
            observations_26,
            environments_26,
            observations_30,
            environments_30,
        ):
            # For 26°
            data = Data()
            data.genotype = [
                {
                    "id": genotype["genotype_id"],
                    "intervention": "deletion",
                    "id_full": genotype["Strain ID"],
                }
            ]
            data.phenotype = {"observation": obs_26, "environment": env_26}
            data_list.append(data)

            # For 30°
            data = Data()
            data.genotype = [
                {
                    "id": genotype["genotype_id"],
                    "intervention": "deletion",
                    "id_full": genotype["Strain ID"],
                }
            ]
            data.phenotype = {"observation": obs_30, "environment": env_30}
            data_list.append(data)

        # Initialize LMDB environment
        log.info("lmdb begin")
        env = lmdb.open(osp.join(self.processed_dir, "data.lmdb"), map_size=int(1e12))

        with env.begin(write=True) as txn:
            # Iterate through each data item
            for idx, item in tqdm(enumerate(data_list)):
                serialized_data = pickle.dumps(item)
                txn.put(f"{idx}".encode(), serialized_data)

        # cache gene property
        self.gene_set = self.compute_gene_set(data_list)

    def preprocess_raw(self, all_data_df: pd.DataFrame, preprocess: dict | None = None):
        # We use the 'Systematic gene name' column as 'genotype' directly
        all_data_df["genotype"] = all_data_df["Systematic gene name"]

        # Find duplicate genotypes
        duplicate_genotypes = all_data_df["genotype"].duplicated(keep=False)
        duplicates_df = all_data_df[duplicate_genotypes].copy()

        # Select which duplicate to keep based on preprocess
        if preprocess is None:
            idx_to_keep = duplicates_df.index
        elif preprocess.get("duplicate_resolution") == "low_std_30":
            idx_to_keep = (
                duplicates_df.dropna(subset=["Single mutant fitness (30°) stddev"])
                .groupby("genotype")["Single mutant fitness (30°) stddev"]
                .idxmin()
            )
        elif preprocess.get("duplicate_resolution") == "high_fitness_30":
            idx_to_keep = (
                duplicates_df.dropna(subset=["Single mutant fitness (30°)"])
                .groupby("genotype")["Single mutant fitness (30°)"]
                .idxmax()
            )
        elif preprocess.get("duplicate_resolution") == "low_fitness_30":
            idx_to_keep = (
                duplicates_df.dropna(subset=["Single mutant fitness (30°)"])
                .groupby("genotype")["Single mutant fitness (30°)"]
                .idxmin()
            )
        elif preprocess.get("duplicate_resolution") == "low_std_26":
            idx_to_keep = (
                duplicates_df.dropna(subset=["Single mutant fitness (26°) stddev"])
                .groupby("genotype")["Single mutant fitness (26°) stddev"]
                .idxmin()
            )
        elif preprocess.get("duplicate_resolution") == "high_fitness_26":
            idx_to_keep = (
                duplicates_df.dropna(subset=["Single mutant fitness (26°)"])
                .groupby("genotype")["Single mutant fitness (26°)"]
                .idxmax()
            )
        elif preprocess.get("duplicate_resolution") == "low_fitness_26":
            idx_to_keep = (
                duplicates_df.dropna(subset=["Single mutant fitness (26°)"])
                .groupby("genotype")["Single mutant fitness (26°)"]
                .idxmin()
            )
        # HACK both abuses the idea of duplicates
        elif preprocess.get("duplicate_resolution") == "low_std_both":
            idx26 = (
                duplicates_df.dropna(subset=["Single mutant fitness (26°) stddev"])
                .groupby("genotype")["Single mutant fitness (26°) stddev"]
                .idxmin()
            )
            idx30 = (
                duplicates_df.dropna(subset=["Single mutant fitness (30°) stddev"])
                .groupby("genotype")["Single mutant fitness (30°) stddev"]
                .idxmin()
            )
            idx_to_keep = list(set(idx26).union(set(idx30)))
        # HACK both abuses the idea of duplicates
        elif preprocess.get("duplicate_resolution") == "low_fitness_both":
            idx26 = (
                duplicates_df.dropna(subset=["Single mutant fitness (26°)"])
                .groupby("genotype")["Single mutant fitness (26°)"]
                .idxmin()
            )
            idx30 = (
                duplicates_df.dropna(subset=["Single mutant fitness (30°)"])
                .groupby("genotype")["Single mutant fitness (30°)"]
                .idxmin()
            )
            idx_to_keep = idx26.append(idx30).unique()
        # HACK both abuses the idea of duplicates
        elif preprocess.get("duplicate_resolution") == "high_fitness_both":
            idx26 = (
                duplicates_df.dropna(subset=["Single mutant fitness (26°)"])
                .groupby("genotype")["Single mutant fitness (26°)"]
                .idxmax()
            )
            idx30 = (
                duplicates_df.dropna(subset=["Single mutant fitness (30°)"])
                .groupby("genotype")["Single mutant fitness (30°)"]
                .idxmax()
            )
            idx_to_keep = idx26.append(idx30).unique()
        else:
            raise ValueError("Unknown preprocess")

        # Drop duplicates, keeping only the selected rows
        duplicates_df = duplicates_df.loc[idx_to_keep]

        # Combine the non-duplicate and selected duplicate rows
        non_duplicates_df = all_data_df[~duplicate_genotypes]
        final_df = pd.concat([non_duplicates_df, duplicates_df], ignore_index=True)

        return final_df

    # New method to save preprocess configuration to a JSON file
    def save_preprocess_config(self, preprocess):
        if not osp.exists(self.preprocess_dir):
            os.makedirs(self.preprocess_dir)
        with open(osp.join(self.preprocess_dir, "preprocess_config.json"), "w") as f:
            json.dump(preprocess, f)

    # New method to load existing preprocess configuration
    def load_preprocess_config(self):
        config_path = osp.join(self.preprocess_dir, "preprocess_config.json")

        if osp.exists(config_path):
            with open(config_path) as f:
                config = json.load(f)
            return config
        else:
            return None

    def len(self) -> int:
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            length = txn.stat()["entries"]

        # Must be closed for dataloader num_workers > 0
        self.close_lmdb()

        return length

    def get(self, idx):
        """Initialize LMDB if it hasn't been initialized yet."""
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None
            data = pickle.loads(serialized_data)
            if self.transform:
                data = self.transform(data)
            return data

    @staticmethod
    def compute_gene_set(data_list):
        computed_gene_set = GeneSet()
        for data in data_list:
            for genotype in data.genotype:
                computed_gene_set.add(genotype["id"])
        return computed_gene_set

    # Reading from JSON and setting it to self._gene_set
    @property
    def gene_set(self):
        try:
            if osp.exists(osp.join(self.preprocess_dir, "gene_set.json")):
                with open(osp.join(self.preprocess_dir, "gene_set.json")) as f:
                    self._gene_set = GeneSet(json.load(f))
            elif self._gene_set is None:
                raise ValueError(
                    "gene_set not written during process. "
                    "Please call compute_gene_set in process."
                )
            return self._gene_set
        # CHECK can probably remove this
        except json.JSONDecodeError:
            raise ValueError("Invalid or empty JSON file found.")

    @gene_set.setter
    def gene_set(self, value):
        if not value:
            raise ValueError("Cannot set an empty or None value for gene_set")
        with open(osp.join(self.preprocess_dir, "gene_set.json"), "w") as f:
            json.dump(list(sorted(value)), f, indent=0)
        self._gene_set = value

    def __repr__(self):
        return f"{self.__class__.__name__}({len(self)})"


class DatasetConfig(ABC):
    @abstractproperty
    def resolution_order(self) -> list[str]:
        pass


class DmfCostanzo2016Config(ModelStrict, DatasetConfig):
    duplicate_resolution: Literal["low_dmf_std", "high_dmf", "low_dmf"] | None = None

    @property
    def resolution_order(self):
        return [str(self.duplicate_resolution)]


class DmfCostanzo2016Dataset(Dataset):
    url = (
        "https://thecellmap.org/costanzo2016/data_files/"
        "Raw%20genetic%20interaction%20datasets:%20Pair-wise%20interaction%20format.zip"
    )

    def __init__(
        self,
        root: str = "data/scerevisiae/costanzo2016",
        subset_n: int = None,
        preprocess: dict | None = None,
        skip_process_file_exist: bool = False,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.subset_n = subset_n
        self._skip_process_file_exist = skip_process_file_exist
        # TODO consider moving to a well defined Dataset class
        self.preprocess = preprocess
        # TODO consider moving to Dataset
        self.preprocess_dir = osp.join(root, "preprocess")
        self._length = None
        self._gene_set = None
        self._df = None
        # Check for existing preprocess config
        existing_config = self.load_preprocess_config()
        if existing_config is not None:
            if existing_config != self.preprocess:
                raise ValueError(
                    "New preprocess does not match existing config."
                    "Delete the processed and process dir for a new Dataset."
                    "Or define a new root."
                )
        super().__init__(root, transform, pre_transform)
        self.env = None

    @property
    def skip_process_file_exist(self):
        return self._skip_process_file_exist

    @property
    def raw_file_names(self) -> list[str]:
        return ["SGA_DAmP.txt", "SGA_ExE.txt", "SGA_ExN_NxE.txt", "SGA_NxN.txt"]

    @property
    def processed_file_names(self) -> list[str]:
        return "data.lmdb"

    def download(self):
        path = download_url(self.url, self.raw_dir)
        with zipfile.ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(self.raw_dir)
        os.remove(path)

        # Move the contents of the subdirectory to the parent raw directory
        sub_dir = os.path.join(
            self.raw_dir,
            "Data File S1. Raw genetic interaction datasets: Pair-wise interaction format",
        )
        for filename in os.listdir(sub_dir):
            shutil.move(os.path.join(sub_dir, filename), self.raw_dir)
        os.rmdir(sub_dir)

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.processed_dir, "data.lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
        )

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None

    @property
    def df(self):
        if osp.exists(osp.join(self.preprocess_dir, "data.csv")):
            self._df = pd.read_csv(osp.join(self.preprocess_dir, "data.csv"))
        return self._df

    @property
    def wt(self):
        wt = {}
        wt["genotype"] = ({"id": None, "intervention": None, "id_full": None},)
        wt["phenotype"] = {
            "observation": {"fitness": 1.0, "genetic_interaction_score": 0},
            "environment": {"media": "YPD", "temperature": 30},
        }
        data = Data()
        data.genotype = wt["genotype"]
        data.phenotype = wt["phenotype"]
        return data

    def process(self):
        os.makedirs(self.preprocess_dir, exist_ok=True)
        self._length = None
        # Initialize an empty DataFrame to hold all raw data
        all_data_df = pd.DataFrame()

        # Read and concatenate all raw files
        print("Reading and Concatenating Raw Files...")
        for file_name in tqdm(self.raw_file_names):
            file_path = os.path.join(self.raw_dir, file_name)

            # Reading data using Pandas; limit rows for demonstration
            df = pd.read_csv(file_path, sep="\t")

            # Concatenating data frames
            all_data_df = pd.concat([all_data_df, df], ignore_index=True)
        # Functions for data filtering... duplicates selection,
        all_data_df = self.preprocess_raw(all_data_df, self.preprocess)
        self.save_preprocess_config(self.preprocess)

        # Subset
        if self.subset_n is not None:
            all_data_df = all_data_df.sample(
                n=self.subset_n, random_state=42
            ).reset_index(drop=True)

        # Save preprocssed df - mainly for quick stats
        all_data_df.to_csv(osp.join(self.preprocess_dir, "data.csv"), index=False)

        print("Processing DMF Files...")

        # Extract genotype information using Polars syntax
        query_genotype = [
            {"id": id_val, "intervention": "deletion", "id_full": full_id}
            for id_val, full_id in zip(
                all_data_df["Query Gene"], all_data_df["Query Strain ID"]
            )
        ]
        array_genotype = [
            {"id": id_val, "intervention": "deletion", "id_full": full_id}
            for id_val, full_id in zip(
                all_data_df["Array Gene"], all_data_df["Array Strain ID"]
            )
        ]

        # Combine the genotypes
        combined_genotypes = list(zip(query_genotype, array_genotype))

        # Extract observation information
        # This part is still a loop due to the complexity of the data structure
        observations = [
            {
                "smf": [row["Query single mutant fitness (SMF)"], row["Array SMF"]],
                "dmf": row["Double mutant fitness"],
                "dmf_std": row["Double mutant fitness standard deviation"],
                "genetic_interaction_score": row["Genetic interaction score (ε)"],
                "genetic_interaction_p-value": row["P-value"],
            }
            for index, row in all_data_df.iterrows()
        ]

        # Create environment dict
        environment = {"media": "YPD", "temperature": 30}

        # Combine everything
        combined_data = [
            {
                "genotype": genotype,
                "phenotype": {"observation": observation, "environment": environment},
            }
            for genotype, observation in zip(combined_genotypes, observations)
        ]

        data_list = []
        for idx, item in tqdm(enumerate(combined_data)):
            data = Data()
            data.genotype = item["genotype"]
            data.phenotype = item["phenotype"]
            data_list.append(data)  # fill the data_list

        # Initialize LMDB environment
        log.info("lmdb begin")
        # TODO make map_size size of disk partition, only virtual address space.
        env = lmdb.open(osp.join(self.processed_dir, "data.lmdb"), map_size=int(1e12))

        # Open a new write transaction
        with env.begin(write=True) as txn:
            # Iterate through each data item
            # TODO loop data_list instead...
            for idx, item in tqdm(enumerate(data_list)):
                serialized_data = pickle.dumps(item)
                txn.put(f"{idx}".encode(), serialized_data)

        # cache gene property
        self.gene_set = self.compute_gene_set(data_list)

    def preprocess_raw(self, all_data_df: pd.DataFrame, preprocess: dict | None = None):
        # Function to extract gene name
        def extract_gene_name(x):
            return x.apply(lambda y: y.split("_")[0])

        # Extract gene names
        query_gene = extract_gene_name(all_data_df["Query Strain ID"]).rename(
            "Query Gene"
        )
        array_gene = extract_gene_name(all_data_df["Array Strain ID"]).rename(
            "Array Gene"
        )

        # Create DataFrame with extracted gene names
        new_df = pd.concat([all_data_df, query_gene, array_gene], axis=1)

        # Function to create and sort genotype
        def create_and_sort_genotype(row):
            query, array = row["Query Gene"], row["Array Gene"]
            return "_".join(sorted([query, array]))

        # Add the genotype column
        new_df["genotype"] = new_df.apply(create_and_sort_genotype, axis=1)

        # Find duplicate genotypes
        duplicate_genotypes = new_df["genotype"].duplicated(keep=False)
        duplicates_df = new_df[duplicate_genotypes].copy()

        # Select which duplicate to keep based on preprocess
        if preprocess.get("duplicate_resolution") == "low_dmf_std":
            # Keep the row with the lowest 'Double mutant fitness standard deviation'
            idx_to_keep = duplicates_df.groupby("genotype")[
                "Double mutant fitness standard deviation"
            ].idxmin()
        elif preprocess.get("duplicate_resolution") == "high_dmf":
            # Keep the row with the highest 'Double mutant fitness'
            idx_to_keep = duplicates_df.groupby("genotype")[
                "Double mutant fitness"
            ].idxmax()
        elif preprocess.get("duplicate_resolution") == "low_dmf":
            # Keep the row with the lowest 'Double mutant fitness'
            idx_to_keep = duplicates_df.groupby("genotype")[
                "Double mutant fitness"
            ].idxmin()
        else:
            raise ValueError("Unknown preprocess")

        # Drop duplicates, keeping only the selected rows
        duplicates_df = duplicates_df.loc[idx_to_keep]

        # Combine the non-duplicate and selected duplicate rows
        non_duplicates_df = new_df[~duplicate_genotypes]
        final_df = pd.concat([non_duplicates_df, duplicates_df], ignore_index=True)

        return final_df

    # New method to save preprocess configuration to a JSON file
    def save_preprocess_config(self, preprocess):
        if not osp.exists(self.preprocess_dir):
            os.makedirs(self.preprocess_dir)
        with open(osp.join(self.preprocess_dir, "preprocess_config.json"), "w") as f:
            json.dump(preprocess, f)

    # TODO implement key merge
    # criterion for merge is defined as key, value is the data object itself.

    # New method to load existing preprocess configuration
    def load_preprocess_config(self):
        config_path = osp.join(self.preprocess_dir, "preprocess_config.json")

        if osp.exists(config_path):
            with open(config_path) as f:
                config = json.load(f)
            return config
        else:
            return None

    def len(self) -> int:
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            length = txn.stat()["entries"]

        # Must be closed for dataloader num_workers > 0
        self.close_lmdb()

        return length

    def get(self, idx):
        """Initialize LMDB if it hasn't been initialized yet."""
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None
            data = pickle.loads(serialized_data)
            if self.transform:
                data = self.transform(data)
            return data

    @staticmethod
    def compute_gene_set(data_list):
        computed_gene_set = GeneSet()
        for data in data_list:
            for genotype in data.genotype:
                computed_gene_set.add(genotype["id"])
        return computed_gene_set

    # Reading from JSON and setting it to self._gene_set
    @property
    def gene_set(self):
        try:
            if osp.exists(osp.join(self.preprocess_dir, "gene_set.json")):
                with open(osp.join(self.preprocess_dir, "gene_set.json")) as f:
                    self._gene_set = GeneSet(json.load(f))
            elif self._gene_set is None:
                raise ValueError(
                    "gene_set not written during process. "
                    "Please call compute_gene_set in process."
                )
            return self._gene_set
        # CHECK can probably remove this
        except json.JSONDecodeError:
            raise ValueError("Invalid or empty JSON file found.")

    @gene_set.setter
    def gene_set(self, value):
        if not value:
            raise ValueError("Cannot set an empty or None value for gene_set")
        with open(osp.join(self.preprocess_dir, "gene_set.json"), "w") as f:
            json.dump(list(sorted(value)), f, indent=0)
        self._gene_set = value

    def __repr__(self):
        return f"{self.__class__.__name__}({len(self)})"


# @prof_input
def main():
    from dotenv import load_dotenv

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    dmf_dataset = DmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e5"),
        preprocess={"duplicate_resolution": "low_dmf_std"},
        # subset_n=100,
    )
    print(dmf_dataset[0])
    print(len(dmf_dataset.gene_set))

    smf_dataset = SmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_smf"),
        preprocess={"duplicate_resolution": "low_std_both"},
        skip_process_file_exist_check=True,
        # subset_n=100,
    )

    print(smf_dataset)
    gene_set = smf_dataset.gene_set.union(dmf_dataset.gene_set)
    print(smf_dataset[0])
    print(len(smf_dataset.gene_set))


if __name__ == "__main__":
    # Load workspace
    main()
    # from dotenv import load_dotenv

    # load_dotenv()
    # DATA_ROOT = os.getenv("DATA_ROOT")
    # os.makedirs(osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016"), exist_ok=True)
    # os.makedirs(
    #     osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016/large"), exist_ok=True
    # )
    # Process data

    # smf_dataset = SMFCostanzo2016Dataset()
    # print(smf_dataset)
    # print(smf_dataset[0])
    # print(len(smf_dataset.gene_set))

    # DMF Small
    # dmf_dataset = DMFCostanzo2016SmallDataset(
    #     root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016")
    # )
    # print(dmf_dataset)
    # print(dmf_dataset[0])

    # dmf_dataset = DMFCostanzo2016Dataset()
    # print(dmf_dataset)
    # print(dmf_dataset[0])
    # print()

    # dmf_dataset_large = DMFCostanzo2016Dataset(
    #     root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_large_nothread")
    # )
    # dmf_dataset_large = DMFCostanzo2016Dataset(
    #     root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016"),
    #     preprocess="low_dmf_std",
    # )
    # print(dmf_dataset_large)
    # print(dmf_dataset_large[0])
    # print(dmf_dataset_large[1]



# torchcell/trainers/graph_convolution_regression.py
# [[torchcell.trainers.graph_convolution_regression]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/trainers/graph_convolution_regression.py
# Test file: torchcell/trainers/test_graph_convolution_regression.py

import math

import lightning as L
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch_geometric.data import Batch, Data
from torchmetrics import (
    MeanAbsoluteError,
    MeanSquaredError,
    MetricCollection,
    PearsonCorrCoef,
    SpearmanCorrCoef,
)
from tqdm import tqdm

import wandb
from torchcell.losses import WeightedMSELoss
from torchcell.viz import fitness, genetic_interaction_score

import torchcell
style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

class GraphConvRegressionTask(L.LightningModule):
    """LightningModule for training models on graph-based regression datasets."""

    def __init__(
        self,
        models: dict[str, nn.Module],
        wt: Data,
        target: str,
        wt_train_per_epoch: float = 10,
        boxplot_every_n_epochs: int = 10,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-5,
        loss: str = "mse",
        batch_size: int = None,
        train_wt_node_loss: bool = False,
        train_epoch_size: int = None,
        clip_grad_norm: bool = False,
        clip_grad_norm_max_norm: float = 0.1,
        order_penalty: bool = False,
        lambda_order: float = 1.0,
        train_mode: bool = True,
        **kwargs,
    ):
        super().__init__()

        # target for training
        self.target = target

        # Lightning settings, doing this for WT embedding
        self.automatic_optimization = False

        self.model_cell = models["cell"]
        self.model_readout = models["readout"]
        self.wt = wt
        self.wt_train_per_epoch = wt_train_per_epoch
        self.is_wt_init = False
        self.wt_nodes_hat, self.wt_set_hat, self.wt_global_hat = None, None, None

        # clip grad norm
        self.clip_grad_norm = clip_grad_norm
        self.clip_grad_norm_max_norm = clip_grad_norm_max_norm

        # order loss
        self.order_penalty = order_penalty
        self.lambda_order = lambda_order

        # train wt diff
        self.train_mode = train_mode
        if self.train_mode == "wt_diff":
            self.x_name = "x"
            self.x_batch_name = "batch"
        elif self.train_mode == "pert":
            self.x_name = "x_pert"
            self.x_batch_name = "x_pert_batch"
        elif self.train_mode == "one_hop_pert":
            # HACK
            # self.x_name = "x_one_hop_pert"
            self.x_name = "x"
            # self.x_batch_name = "x_one_hop_pert_batch"
            self.x_batch_name = "x_batch"

        # loss
        if loss == "mse":
            self.loss = nn.MSELoss()
        elif loss == "weighted_mse":
            mean_value = kwargs.get("fitness_mean_value")
            penalty = kwargs.get("penalty", 1.0)
            self.loss = WeightedMSELoss(mean_value=mean_value, penalty=penalty)
        elif loss == "mae":
            self.loss = nn.L1Loss()
        else:
            raise ValueError(
                f"Loss type '{loss}' is not valid."
                "Currently, supports 'mse' or 'mae' loss."
            )
        self.loss_node = nn.MSELoss()

        # Node Loss Regularization to Unit Vector
        self.train_wt_node_loss = train_wt_node_loss

        # train epoch size for wt frequency
        self.train_epoch_size = train_epoch_size

        # optimizer
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

        # batch_size
        self.batch_size = batch_size

        self.train_metrics = MetricCollection(
            {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
            },
            prefix="train_",
        )
        self.val_metrics = self.train_metrics.clone(prefix="val_")
        self.test_metrics = self.train_metrics.clone(prefix="test_")

        # Separate attributes for Pearson and Spearman correlation coefficients
        self.pearson_corr = PearsonCorrCoef()
        self.spearman_corr = SpearmanCorrCoef()

        # Used in end for whisker plot
        self.boxplot_every_n_epochs = boxplot_every_n_epochs
        self.true_values = {}
        self.predictions = {}
        for target in self.target:
            self.true_values[target] = []
            self.predictions[target] = []

        # wandb model artifact logging
        self.last_logged_best_step = None

    def setup(self, stage=None):
        self.model_cell = self.model_cell.to(self.device)
        self.model_readout = self.model_readout.to(self.device)

    def forward(self, x, batch, edge_index):
        inst_nodes_hat, inst_set_hat = self.model_cell(x, batch, edge_index)
        # This case is only for sanity checking
        if self.wt_set_hat is None and self.train_mode == "wt_diff":
            self.wt_set_hat = torch.ones_like(inst_set_hat)
        if self.train_mode == "wt_diff":
            y_set_hat = self.wt_set_hat.mean(dim=0) - inst_set_hat
        else:
            y_set_hat = inst_set_hat
        y_hat = self.model_readout(y_set_hat)
        return y_hat

    def on_train_start(self):
        # Calculate the model size (number of parameters)
        parameter_size = sum(p.numel() for p in self.parameters())
        # Log it using wandb
        self.log("model/parameters_size", parameter_size)

    def train_wt(self):
        # CHECK on definition of global_step - refresh with epoch?
        if self.global_step == 0 and not self.is_wt_init:
            wt_batch = Batch.from_data_list([self.wt] * self.batch_size).to(self.device)
            self.wt_nodes_hat, self.wt_set_hat = self.model_cell(
                wt_batch.x, wt_batch.batch
            )

            self.is_wt_init = True
        if (self.global_step == 0) or self.global_step % math.ceil(
            (self.train_epoch_size + 1) / self.wt_train_per_epoch
        ) == 0:
            wt_y_hats = []
            loss_wts = []

            progress_bar = tqdm(desc="Processing", position=0, leave=True)

            wt_y_hat_mean = np.nan
            while True:
                # Global Loss
                # set up optimizer
                opt = self.optimizers()
                opt.zero_grad()

                wt_batch = Batch.from_data_list([self.wt] * self.batch_size).to(
                    self.device
                )
                self.wt_y_hat = self(wt_batch.x, wt_batch.batch)
                loss_wt = self.loss(self.wt_y_hat, wt_batch[self.target])
                self.log("wt loss", loss_wt)
                self.log("wt mean", self.wt_y_hat.detach().mean())
                self.log("wt batch fitness mean", wt_batch[self.target].mean())
                # get updated wt reference
                if self.train_wt_node_loss:
                    self.wt_nodes_hat, self.wt_set_hat = self.model_cell(
                        wt_batch.x, wt_batch.batch
                    )
                    # Node Loss
                    loss_nodes = self.loss_node(
                        self.wt_nodes_hat, torch.ones_like(self.wt_nodes_hat)
                    )
                    self.log("wt loss_nodes", loss_nodes)
                    self.manual_backward(loss_wt + loss_nodes)
                else:
                    self.manual_backward(loss_wt)

                if self.clip_grad_norm:
                    nn.utils.clip_grad_norm_(
                        self.parameters(), max_norm=self.clip_grad_norm_max_norm
                    )
                # finish optimization
                opt.step()
                opt.zero_grad()

                # Get updated wt reference
                self.model_cell.eval()
                with torch.no_grad():
                    self.wt_nodes_hat, self.wt_set_hat = self.model_cell(
                        wt_batch.x, wt_batch.batch
                    )
                self.model_cell.train()
                ###
                wt_y_hat_mean = self.wt_y_hat.mean().cpu().detach().numpy()
                wt_y_hats.append(wt_y_hat_mean)
                loss_wts.append(loss_wt.cpu().detach().numpy())
                progress_bar.update(1)
                if 0.99 < wt_y_hat_mean < 1.01 or self.current_epoch < 10:
                    # if self.current_epoch >= 2:
                    #     plt.plot(wt_y_hats)
                    #     plt.show()
                    #     plt.plot(loss_wts)
                    #     plt.show()
                    break
            progress_bar.close()

    def ordering_penalty(self, y_hat, y):
        # Step 1: Obtain the indices that would sort y
        _, sorted_indices = torch.sort(y, descending=False)

        # Step 2: Use these indices to rearrange y_hat
        y_hat_sorted = torch.gather(y_hat, 0, sorted_indices)

        # Step 3: Compute the penalty based on rearranged y_hat
        diffs = y_hat_sorted[1:] - y_hat_sorted[:-1]
        violations = torch.clamp(diffs, max=0)
        penalty = -torch.sum(violations)

        return penalty

    def training_step(self, batch, batch_idx):
        # Train on wt reference
        if self.train_mode == "wt_diff":
            self.train_wt()

        y = self.compose_target(batch)
        y_hat = self(
            x=batch[self.x_name],
            batch=batch[self.x_batch_name],
            edge_index=batch["edge_index"],
        )
        opt = self.optimizers()
        opt.zero_grad()
        assert y_hat.size() == y.size(), "y_hat and y should have the same size"

        losses = {}
        batch_size = batch[self.x_batch_name][-1].item() + 1

        # Calculate loss for each target
        for target in self.target:
            i = self.target.index(target)
            y_hat_target = y_hat[:, i]
            y_target = y[:, i]
            if target == "genetic_interaction_score":
                y_hat_target = 4.20 * y_hat_target
                y_target = 4.20 * y_target
                losses[target] = self.loss(y_hat_target, y_target)
                y_hat_target = 4.20 * y_hat_target
                y_target = 4.20 * y_target
            elif target == "fitness":
                losses[target] = self.loss(y_hat_target, y_target)
            self.log(
                f"{target}/train_loss",
                losses[target],
                batch_size=batch_size,
                sync_dist=True,
            )
            self.train_metrics(y_hat_target, y_target)
            self.log(
                f"{target}/train_pearson",
                self.pearson_corr(y_hat_target, y_target),
                batch_size=batch_size,
                sync_dist=True,
            )
            self.log(
                f"{target}/train_spearman",
                self.spearman_corr(y_hat_target, y_target),
                batch_size=batch_size,
                sync_dist=True,
            )

        # Overall loss
        if self.order_penalty:
            order_penalty_value = self.ordering_penalty(y_hat, y)
            order_penalty_loss = self.lambda_order * order_penalty_value
            self.log("order_penalty_loss", order_penalty_loss)
            loss = sum(losses.values()) + self.lambda_order * order_penalty_loss
        else:
            loss = sum(losses.values())

        self.manual_backward(loss)
        if self.clip_grad_norm:
            nn.utils.clip_grad_norm_(
                self.parameters(), max_norm=self.clip_grad_norm_max_norm
            )
        opt.step()
        opt.zero_grad()

        # Log overall metrics
        self.log("train_loss", loss, batch_size=batch_size, sync_dist=True)
        self.train_metrics(y_hat, y)

        return loss

    def on_train_epoch_end(self):
        self.log_dict(self.train_metrics.compute(), sync_dist=True)
        self.train_metrics.reset()

    def compose_target(self, batch):
        composed_target = torch.tensor([], dtype=torch.float32, device=self.device)
        for target in self.target:
            composed_target = torch.cat(
                [composed_target, batch.get(target).unsqueeze(1)], dim=1
            )
        if len(self.target) == 1:
            composed_target = composed_target.squeeze(-1)
        return composed_target

    def validation_step(self, batch, batch_idx):
        # Extract the batch vector
        y = self.compose_target(batch)
        y_hat = self(
            x=batch[self.x_name],
            batch=batch[self.x_batch_name],
            edge_index=batch["edge_index"],
        )
        losses = {}
        batch_size = batch[self.x_batch_name][-1].item() + 1
        for target in self.target:
            i = self.target.index(target)
            y_hat_target = y_hat[:, i]
            y_target = y[:, i]
            if target == "genetic_interaction_score":
                y_hat_target = 4.20 * y_hat_target
                y_target = 4.20 * y_target
                losses[target] = self.loss(y_hat_target, y_target)
                y_hat_target = 4.20 * y_hat_target
                y_target = 4.20 * y_target
            elif target == "fitness":
                losses[target] = self.loss(y_hat_target, y_target)
            self.log(
                f"{target}/val_loss",
                losses[target],
                batch_size=batch_size,
                sync_dist=True,
            )
            self.val_metrics(y_hat_target, y_target)
            self.log(
                f"{target}/val_pearson",
                self.pearson_corr(y_hat_target, y_target),
                batch_size=batch_size,
                sync_dist=True,
            )
            self.log(
                f"{target}/val_spearman",
                self.spearman_corr(y_hat_target, y_target),
                batch_size=batch_size,
                sync_dist=True,
            )
        loss = sum(losses.values())
        self.log("val_loss", loss, batch_size=batch_size, sync_dist=True)
        self.val_metrics(y_hat, y)
        # for i in enumerate(self.target):
        # Logging the correlation coefficients
        for target in self.target:
            i = self.target.index(target)
            self.true_values[target].append(y_target.detach())
            self.predictions[target].append(y_hat_target.detach())

    def on_validation_epoch_end(self):
        self.log_dict(self.val_metrics.compute(), sync_dist=True)
        self.val_metrics.reset()

        # Skip plotting during sanity check
        if self.trainer.sanity_checking or (
            self.current_epoch % self.boxplot_every_n_epochs != 0
        ):
            return

        # Convert lists to tensors
        for target in self.target:
            target_true_value = torch.cat(self.true_values[target], dim=0)
            target_predictions = torch.cat(self.predictions[target], dim=0)

            if target == "fitness":
                fig = fitness.box_plot(target_true_value, target_predictions)
            if target == "genetic_interaction_score":
                fig = genetic_interaction_score.box_plot(
                    target_true_value, target_predictions
                )
            wandb.log({"binned_values_box_plot": wandb.Image(fig)})
            plt.close(fig)
            # Clear the stored values for the next epoch
            self.true_values[target] = []
            self.predictions[target] = []

            current_global_step = self.global_step
            if (
                self.trainer.checkpoint_callback.best_model_path
                and current_global_step != self.last_logged_best_step
            ):
                # Save model as a W&B artifact
                artifact = wandb.Artifact(
                    name=f"model-global_step-{current_global_step}",
                    type="model",
                    description=f"Model on validation epoch end step - {current_global_step}",
                    metadata=dict(self.hparams),
                )
                artifact.add_file(self.trainer.checkpoint_callback.best_model_path)
                wandb.log_artifact(artifact)
                self.last_logged_best_step = (
                    current_global_step  # update the last logged step
                )

    def test_step(self, batch, batch_idx):
        # Extract the batch vector
        y = self.compose_target(batch)
        y_hat = self(
            x=batch[self.x_name],
            batch=batch[self.x_batch_name],
            edge_index=batch["edge_index"],
        )
        loss = self.loss(y_hat, y)
        batch_size = batch[self.x_batch_name][-1].item() + 1
        self.log("test_loss", loss, batch_size=batch_size, sync_dist=True)
        self.test_metrics(y_hat, y)
        # Logging the correlation coefficients
        self.log(
            "test_pearson",
            self.pearson_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "test_spearman",
            self.spearman_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )

    def on_test_epoch_end(self):
        self.log_dict(self.test_metrics.compute(), sync_dist=True)
        self.test_metrics.reset()

    def configure_optimizers(self):
        params = list(self.model_cell.parameters()) + list(
            self.model_readout.parameters()
        )
        optimizer = torch.optim.Adam(
            params, lr=self.learning_rate, weight_decay=self.weight_decay
        )
        return optimizer



# torchcell/multidigraph/sgd.py
# [[torchcell.multidigraph.sgd]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/multidigraph/sgd.py
# Test file: torchcell/multidigraph/test_sgd.py
import asyncio
import json
import logging
import os
import os.path as osp
import tempfile
import time
from asyncio import Task
from collections.abc import Callable
from typing import Any, Optional

import aiohttp
from aiohttp import ClientError, ContentTypeError
from attrs import define, field
from tqdm import tqdm

from torchcell.multidigraph.validation.locus_related.locus import (
    Alias,
    InteractionOverview,
    LocusData,
    LocusDataUrl,
    PhysicalExperiments,
    Qualities,
    Reference,
    validate_data,
)

log = logging.getLogger(__name__)


@define
class Gene:
    locusID: str = "YAL001C"
    is_validated: bool = field(default=True, init=True, repr=False)
    sgd_url: str = "https://www.yeastgenome.org/backend/locus"
    headers: dict[str, str] = field(default={"accept": "application/json"})
    base_data_dir: str = "data/sgd/genome/genes"
    save_path: str = field(default=None, init=False, repr=True)
    _data: dict[str, dict[Any, Any] | list[Any]] = field(factory=dict, init=False)
    _data_task: Task[Any] | None = field(default=None, init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        if not osp.exists(self.base_data_dir):
            os.makedirs(self.base_data_dir)
        self.save_path: str = osp.join(self.base_data_dir, f"{self.locusID}.json")
        if osp.exists(self.save_path):
            self._data = self.read()
        else:
            # No fetch operation scheduled yet
            self._data_task = None

    @property
    def data(self) -> dict[str, dict[Any, Any] | list[Any]]:
        if self._data != {}:
            return self._data
        else:
            raise ValueError(
                "No data available. Call fetch_data(), e.g., `asyncio.run(gene.fetch_data())`"
            )

    async def fetch_data(self) -> None:
        if self._data_task is None:
            # Schedule the download
            self._data_task = asyncio.create_task(self.download_data())
            await self._data_task
        elif not self._data_task.done():
            # Download already scheduled but not yet finished
            await self._data_task
        # If task is done and _data is still empty, download_data failed silently
        if self._data == {}:
            raise ValueError("Data fetch failed.")

    async def download_data(self) -> None:
        self._data["locus"] = await self.locus()
        self._data["sequence_details"] = await self.sequence_details()
        self._data["neighbor_sequence_details"] = await self.neighbor_sequence_details()
        self._data["posttranslational_details"] = await self.posttranslational_details()
        self._data[
            "protein_experiment_details"
        ] = await self.protein_experiment_details()
        self._data["protein_domain_details"] = await self.protein_domain_details()
        self._data["go_details"] = await self.go_details()
        self._data["phenotype_details"] = await self.phenotype_details()
        self._data["interaction_details"] = await self.interaction_details()
        self._data["regulation_details"] = await self.regulation_details()
        self._data["literature_details"] = await self.literature_details()
        self.write()

    def write(self) -> None:
        with open(self.save_path, "w") as f:
            json.dump(self._data, f, indent=4)

    def read(self) -> dict[str, dict[Any, Any] | list[Any]]:
        if not osp.exists(self.save_path):
            raise ValueError(f"File {self.save_path} does not exist")
        with open(self.save_path) as f:
            data_in = json.load(f)
            if not isinstance(data_in, dict):
                raise ValueError(f"File {self.save_path} is not a dict")
        return data_in

    # With error handling
    async def _get_data(
        self, url: str, max_retries: int = 10
    ) -> dict[Any, Any] | list[Any] | None:
        for retry in range(max_retries):
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(url, headers=self.headers) as response:
                        content_type = response.headers.get("Content-Type")

                        # Check if the content type indicates a JSON response
                        if "application/json" in content_type:
                            data = await response.json()
                            if not isinstance(data, (dict, list)):
                                raise ValueError(f"Data is not a dict or list: {data}")
                            return data
                        else:
                            # Save unexpected content (e.g., HTML) to a temporary file for debugging
                            content = await response.text()
                            with tempfile.NamedTemporaryFile(
                                delete=False, suffix=".html"
                            ) as temp:
                                temp.write(content.encode())
                                logging.error(
                                    f"Saved unexpected response to: {temp.name}"
                                )

                            error_message = (
                                f"Unexpected content type: {content_type}. URL: {url}"
                            )
                            logging.error(error_message)
                            raise ContentTypeError(error_message)

            except (ClientError, ContentTypeError) as e:
                if retry == max_retries - 1:  # If it's the last retry and still fails
                    logging.error(
                        f"Failed to fetch data after {max_retries} attempts. Error: {e}"
                    )
                    return None
                else:
                    logging.warning(f"Attempt {retry + 1} failed. Retrying...")
                    await asyncio.sleep(2**retry)  # Exponential backoff

        return None

    async def locus(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID)
        data = await self._get_data(url)
        if self.is_validated:
            data = validate_data(data)
            with open("locus_schema.json", "w") as f:
                json.dump(data.model_json_schema(), f, indent=4)
            data = data.model_dump()
        return data  # breakpoint, printed out data

    async def sequence_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "sequence_details")
        data = await self._get_data(url)
        return data

    async def neighbor_sequence_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "neighbor_sequence_details")
        data = await self._get_data(url)
        return data

    async def posttranslational_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "posttranslational_details")
        data = await self._get_data(url)
        return data

    async def protein_experiment_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "protein_experiment_details")
        data = await self._get_data(url)
        return data

    async def protein_domain_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "protein_domain_details")
        data = await self._get_data(url)
        return data

    async def go_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "go_details")
        data = await self._get_data(url)
        return data

    async def phenotype_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "phenotype_details")
        data = await self._get_data(url)
        return data

    async def interaction_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "interaction_details")
        data = await self._get_data(url)
        return data

    async def regulation_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "regulation_details")
        data = await self._get_data(url)
        return data

    async def literature_details(self) -> dict[Any, Any] | list[Any]:
        url = osp.join(self.sgd_url, self.locusID, "literature_details")
        data = await self._get_data(url)
        return data


async def process_gene(gene: Gene, progress_bar: Any) -> dict[Any, Any] | list[Any]:
    await gene.fetch_data()
    data = gene.data
    progress_bar.update()
    return data


async def download_genes(
    locus_ids: list[str], gene_factory: Callable[[str], Gene], is_validated: bool
) -> None:
    with tqdm(total=len(locus_ids)) as progress_bar:
        tasks = []
        for id_ in locus_ids:
            # Check if the file for the gene already exists
            file_path = f"data/sgd/genes/{id_}.json"

            if os.path.exists(file_path):
                logging.info(f"Data for gene {id_} already exists. Skipping...")
                progress_bar.update(1)
                continue

            gene = gene_factory(id_, is_validated)
            tasks.append(process_gene(gene, progress_bar))

        await asyncio.gather(*tasks)


def create_gene(locusID: str, is_validated: bool) -> Gene:
    return Gene(locusID=locusID, is_validated=is_validated)


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


async def download_gene_chunk(chunk, create_gene_fn, validate_flag):
    """Download a chunk of genes with a delay before starting."""
    await asyncio.sleep(1)  # Give a small break between chunks
    await download_genes(chunk, create_gene_fn, validate_flag)


def main_get_all_genes():
    from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

    genome = SCerevisiaeGenome()
    locus_ids = list(genome.gene_set)

    CHUNK_SIZE = 50  # Adjust this value based on what works best for you
    locus_id_chunks = list(chunks(locus_ids, CHUNK_SIZE))

    for chunk in locus_id_chunks:
        asyncio.run(download_gene_chunk(chunk, create_gene, False))


if __name__ == "__main__":
    main_get_all_genes()
    # main_median_protein()
    # main()


# def main_get_all_genes():
#     from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

#     genome = SCerevisiaeGenome()
#     # genome.drop_chrmt()
#     locus_ids = list(genome.gene_set)

#     # Chunk the locus_ids list by every 10 items
#     # locus_id_chunks = chunks(locus_ids, 10)

#     asyncio.run(download_genes(locus_ids[:100], create_gene, is_validated=False))

# def main_median_protein():
#     import os

#     from dotenv import load_dotenv

#     load_dotenv()
#     DATA_ROOT = os.getenv("DATA_ROOT")
#     from torchcell.datasets.scerevisiae import DmfCostanzo2016Dataset
#     from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

#     dmf_dataset = DmfCostanzo2016Dataset(
#         root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e4"),
#         preprocess="low_dmf_std",
#     )
#     # genome = SCerevisiaeGenome(data_root=osp.join(DATA_ROOT, "data/sgd/genome"))
#     # genome.drop_chrmt()
#     # genome.drop_empty_go()
#     locus_ids = list(dmf_dataset.gene_set)
#     start_time = time.time()
#     asyncio.run(download_genes(locus_ids, create_gene, is_validated=False))
#     end_time = time.time()
#     print(f"Execution time: {end_time - start_time} seconds")

# def main() -> None:
#     start_time = time.time()
#     locus_ids = [
#         "YPR201W",
#         "YPR202W",
#         "YPR199C",
#         "YPR200C",
#         "YPR198W",
#         "YPR196W",
#         "YPR197C",
#         "YPR194C",
#         "YPR195C",
#         "YPR193C",
#         "YPR192W",
#         "YLR153C",
#     ]
#     asyncio.run(
#         download_genes(locus_ids, create_gene, is_validated=False)
#     )  # TODO Change for Dev
#     end_time = time.time()
#     print(f"Execution time: {end_time - start_time} seconds")
#     # gene = Gene("YDR210W")
#     # asyncio.run(gene.fetch_data())
#     # gene.data


# async def download_genes(
#     locus_ids: list[str], gene_factory: Callable[[str], Gene], is_validated: bool
# ) -> None:
#     with tqdm(total=len(locus_ids)) as progress_bar:
#         await asyncio.gather(
#             *(
#                 process_gene(gene_factory(id_, is_validated), progress_bar)
#                 for id_ in locus_ids
#             )
#         )

# async def _get_data(self, url: str) -> dict[Any, Any] | list[Any]:
#     async with aiohttp.ClientSession() as session:
#         async with session.get(url, headers=self.headers) as response:
#             data = await response.json()
#             if not isinstance(data, (dict, list)):
#                 raise ValueError(f"Data is not a dict or list: {data}")
#             return data



# torchcell/trainers/regression_deep_set_transformer.py
# [[torchcell.trainersu.regression_deep_set_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/trainers/regression_deep_set_transformer.py
# Test file: torchcell/trainers/test_regression_deep_set_transformer.py

import lightning as L
import matplotlib.pyplot as plt
import torch
import torch.nn as nn
from torch_geometric.data import Batch, Data
from torchmetrics import (
    MeanAbsoluteError,
    MeanSquaredError,
    MetricCollection,
    PearsonCorrCoef,
    SpearmanCorrCoef,
)

import wandb
from torchcell.losses import WeightedMSELoss

import torchcell
style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

class RegressionTaskDeepSetTransformer(L.LightningModule):
    """LightningModule for training models on graph-based regression datasets."""

    target_key: str = "fitness"

    def __init__(
        self,
        models: dict[str, nn.Module],
        wt: Data,
        wt_train_ratio: int = 10,
        boxplot_every_n_epochs: int = 10,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-5,
        loss: str = "mse",
    ):
        super().__init__()

        # Lightning settings, doing this for WT embedding
        self.automatic_optimization = False

        self.model_dst = models["deep_set_transformer"]
        self.model_lin = models["mlp_ref_set"]
        self.wt = wt
        self.wt_train_ratio = wt_train_ratio
        self.is_wt_init = False
        self.wt_global_hat, self.wt_set_hat, self.wt_nodes_hat = None, None, None

        if loss == "mse":
            self.loss = nn.MSELoss()
        if loss == "weighted_mse":
            self.loss = WeightedMSELoss(mean_value=0.868812)
        elif loss == "mae":
            self.loss = nn.L1Loss()
        else:
            raise ValueError(
                f"Loss type '{loss}' is not valid."
                "Currently, supports 'mse' or 'mae' loss."
            )
        self.loss_node = nn.MSELoss()

        self.l1_lambda = 1e-7  # 0.001

        # optimizer
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

        self.train_metrics = MetricCollection(
            {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
            },
            prefix="train_",
        )
        self.val_metrics = self.train_metrics.clone(prefix="val_")
        self.test_metrics = self.train_metrics.clone(prefix="test_")

        # Separate attributes for Pearson and Spearman correlation coefficients
        self.pearson_corr = PearsonCorrCoef()
        self.spearman_corr = SpearmanCorrCoef()

        # Used in end for whisker plot
        self.boxplot_every_n_epochs = boxplot_every_n_epochs
        self.true_values = []
        self.predictions = []

    def setup(self, stage=None):
        self.model_dst = self.model_dst.to(self.device)
        self.model_lin = self.model_lin.to(self.device)

    def forward(self, x, batch):
        inst_nodes_hat, inst_set_hat, _, attn_weights = self.model_dst(x, batch)
        # This case is only for sanity checking
        if self.wt_set_hat is None:
            self.wt_set_hat = torch.ones_like(inst_set_hat)
        y_set_hat = self.wt_set_hat.mean(dim=0) - inst_set_hat
        y_hat = self.model_lin(y_set_hat)
        return y_hat, attn_weights

    def on_train_start(self):
        # Calculate the model size (number of parameters)
        parameter_size = sum(p.numel() for p in self.parameters())
        # Log it using wandb
        self.log("model/parameters_size", parameter_size)

    def train_wt(self):
        # CHECK on definition of global_step - refresh with epoch?
        if self.global_step == 0 and not self.is_wt_init:
            wt_batch = Batch.from_data_list([self.wt, self.wt]).to(self.device)
            self.wt_nodes_hat, self.wt_set_hat, _, attn_weights = self.model_dst(
                wt_batch.x, wt_batch.batch
            )

            self.is_wt_init = True
        if self.global_step == 0 or self.global_step % self.wt_train_ratio == 0:
            ################ Global
            # set up optimizer
            opt = self.optimizers()
            opt.zero_grad()
            # train on wt
            wt_batch = Batch.from_data_list([self.wt] * 16).to(self.device)
            self.wt_y_hat, attn_weights = self(wt_batch.x, wt_batch.batch)
            # rand_pert = torch.FloatTensor(16).uniform_(1.00001, 1.0001).to(self.device)
            loss_wt = self.loss(self.wt_y_hat, wt_batch.fitness)
            self.log("wt loss", loss_wt)
            self.log("wt mean", self.wt_y_hat.detach().mean())
            # self.manual_backward(loss)  # error on this line
            # nn.utils.clip_grad_norm_(self.parameters(), max_norm=1.0)

            # opt.step()
            opt.zero_grad()
            # Get updated wt reference
            # Set the model to evaluation mode
            # self.model_dst.eval()

            # # get updated wt reference
            # with torch.no_grad():
            self.wt_nodes_hat, self.wt_set_hat, _, attn_weights = self.model_dst(
                wt_batch.x, wt_batch.batch
            )
            # Revert the model back to training mode
            # self.model_dst.train()
            ######### Node
            # opt = self.optimizers()
            # opt.zero_grad()

            loss_nodes = self.loss_node(
                self.wt_nodes_hat, torch.ones_like(self.wt_nodes_hat)
            )
            self.log("wt loss_nodes", loss_nodes)

            total_loss = (
                loss_wt + loss_nodes + self.l1_lambda * attn_weights.abs().sum()
            )
            self.manual_backward(total_loss)
            nn.utils.clip_grad_norm_(self.parameters(), max_norm=1.0)
            opt.step()
            opt.zero_grad()
            # Get updated wt reference
            # Set the model to evaluation mode
            self.model_dst.eval()

            # get updated wt reference
            with torch.no_grad():
                self.wt_nodes_hat, self.wt_set_hat, _, attn_weights = self.model_dst(
                    wt_batch.x, wt_batch.batch
                )
            # Revert the model back to training mode
            self.model_dst.train()
            ########## Break after overfit
            # if self.current_epoch < 2:
            #     break

            # elif loss_global < 1:
            #     print(f"Broke WT overfit on: {i}")
            #     break

    def training_step(self, batch, batch_idx):
        # Train on wt reference
        self.train_wt()
        # Extract the batch vector
        x, y, batch_vector = batch.x, batch.fitness, batch.batch
        # Pass the batch vector to the forward method
        y_hat, attn_weights = self(x, batch_vector)

        opt = self.optimizers()
        opt.zero_grad()
        l1_attn_loss = self.l1_lambda * attn_weights.abs().sum()
        loss = self.loss(y, y_hat) + l1_attn_loss
        self.manual_backward(loss)  # error on this line
        nn.utils.clip_grad_norm_(self.parameters(), max_norm=1.0)
        opt.step()
        opt.zero_grad()
        # logging
        batch_size = batch_vector[-1].item() + 1
        self.log("train_loss", loss, batch_size=batch_size, sync_dist=True)
        self.log(
            "train_loss_l1",
            l1_attn_loss.detach(),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.train_metrics(y_hat, y)
        # Logging the correlation coefficients
        self.log(
            "train_pearson",
            self.pearson_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "train_spearman",
            self.spearman_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        return loss

    def on_train_epoch_end(self):
        self.log_dict(self.train_metrics.compute(), sync_dist=True)
        self.train_metrics.reset()

    def validation_step(self, batch, batch_idx):
        # Extract the batch vector
        x, y, batch_vector = batch.x, batch.fitness, batch.batch
        y_hat, attn_weights = self(x, batch_vector)
        loss = self.loss(y_hat, y)
        batch_size = batch_vector[-1].item() + 1
        self.log("val_loss", loss, batch_size=batch_size, sync_dist=True)
        self.val_metrics(y_hat, y)
        # Logging the correlation coefficients
        self.log(
            "val_pearson",
            self.pearson_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "val_spearman",
            self.spearman_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.true_values.append(y.detach())
        self.predictions.append(y_hat.detach())

    def on_validation_epoch_end(self):
        self.log_dict(self.val_metrics.compute(), sync_dist=True)
        self.val_metrics.reset()

        # Skip plotting during sanity check
        if self.trainer.sanity_checking or (
            self.current_epoch % self.boxplot_every_n_epochs != 0
        ):
            return

        # Convert lists to tensors
        true_values = torch.cat(self.true_values, dim=0)
        predictions = torch.cat(self.predictions, dim=0)

        # Define bins
        bins = [0, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, float("inf")]

        # Bin predictions and collect corresponding true values
        binned_true_values = []
        bin_labels = []
        for i in range(len(bins) - 1):
            mask = (predictions >= bins[i]) & (predictions < bins[i + 1])
            binned_values = true_values[mask].cpu().numpy()
            binned_true_values.append(binned_values)
            bin_labels.append(f"{bins[i]}-{bins[i+1]}")

        # Create a box plot using matplotlib
        fig, ax = plt.subplots()
        ax.boxplot(binned_true_values, labels=bin_labels)
        ax.set_ylabel("True Values")
        ax.set_xlabel("Prediction Bins")
        ax.set_title("Box plot of True Values for each Prediction Bin")

        # Log the plot to wandb
        wandb.log({"binned_values_box_plot": wandb.Image(fig)})

        # Clear the stored values for the next epoch
        self.true_values = []
        self.predictions = []

        # Close the matplotlib plot
        plt.close(fig)

    def test_step(self, batch, batch_idx):
        # Extract the batch vector
        x, y, batch_vector = batch.x, batch.fitness, batch.batch
        y_hat, attn_weights = self(x, batch_vector)
        loss = self.loss(y_hat, y)
        batch_size = batch_vector[-1].item() + 1
        self.log("test_loss", loss, batch_size=batch_size, sync_dist=True)
        self.test_metrics(y_hat, y)
        # Logging the correlation coefficients
        self.log(
            "test_pearson",
            self.pearson_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "test_spearman",
            self.spearman_corr(y_hat, y),
            batch_size=batch_size,
            sync_dist=True,
        )

    def on_test_epoch_end(self):
        self.log_dict(self.test_metrics.compute(), sync_dist=True)
        self.test_metrics.reset()

    def configure_optimizers(self):
        params = list(self.model_dst.parameters()) + list(self.model_lin.parameters())
        optimizer = torch.optim.Adam(
            params, lr=self.learning_rate, weight_decay=self.weight_decay
        )
        return optimizer



# torchcell/datasets/genome.py
# [[torchcell.datasets.genome]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/genome.py
# Test file: torchcell/datasets/test_genome.py

from abc import ABC, abstractmethod

from torch_geometric.data import Data, InMemoryDataset, download_url


class GenomeDataset(InMemoryDataset, ABC):
    pass

    def __add__():
        pass


class SCerevisiaeS88C(GenomeDataset):
    pass


if __name__ == "__main__":
    pass



import lmdb
import os.path as osp
import json
from tqdm import tqdm

def iterate_lmdb_database(lmdb_dir: str):
    """
    Iterates over all key-value pairs in an LMDB database and interprets the values as JSON.

    Parameters:
    lmdb_dir (str): The directory where the LMDB database is stored.
    """
    # Ensure the LMDB directory exists
    if not osp.exists(lmdb_dir):
        print(f"LMDB directory {lmdb_dir} does not exist.")
        return

    # Open the LMDB environment in read mode
    env = lmdb.open(lmdb_dir, readonly=True)
    with env.begin() as txn:
        # Create a cursor to iterate over the key-value pairs
        cursor = txn.cursor()
        for key, value in tqdm(cursor):
            # Deserialize the value from JSON to a Python dict
            value_as_dict = json.loads(value.decode())
            # Here, you can process each key-value pair as needed
            # For demonstration, we'll just print the key and the deserialized value
            print(f"Key: {key.decode()}, Value: {value_as_dict}")

    # Close the LMDB environment
    env.close()

# Example usage
if __name__ == "__main__":
    lmdb_dir = "data/torchcell/dmf-2022_02_12/raw/data.lmdb"  # Adjust the path to your LMDB directory
    iterate_lmdb_database(lmdb_dir)



# torchcell/datasets/nucleotide_transformer.py
# [[torchcell.datasets.nucleotide_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/nucleotide_transformer.py
# Test file: tests/torchcell/datasets/test_nucleotide_transformer.py
import os
from collections.abc import Callable
from typing import Optional

import torch
from torch_geometric.data import Data, InMemoryDataset
from tqdm import tqdm

from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.models.nucleotide_transformer import NucleotideTransformer
from torchcell.sequence import ParsedGenome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

os.makedirs("data/scerevisiae/nucleotide_transformer_embed", exist_ok=True)


# TODO rename utr to up and down
class NucleotideTransformerDataset(BaseEmbeddingDataset):
    MODEL_TO_WINDOW = {
        "nt_window_5979_max": ("window", 5979, True),
        "nt_window_5979": ("window", 5979, False),
        "nt_window_3utr_5979": ("window_3utr", 5979, False),
        "nt_window_3utr_5979_undersize": ("window_3utr", 5979, True),
        "nt_window_5utr_5979": ("window_5utr", 5979, False),
        "nt_window_5utr_5979_undersize": ("window_5utr", 5979, True),
        "nt_window_3utr_300": ("window_3utr", 300, False),
        "nt_window_3utr_300_undersize": ("window_3utr", 300, True),
        "nt_window_5utr_1000": ("window_5utr", 1000, False),
        "nt_window_5utr_1000_undersize": ("window_5utr", 1000, True),
    }

    def __init__(
        self,
        root: str,
        genome: SCerevisiaeGenome,
        model_name: str | None = None,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.genome = genome
        super().__init__(root, transform, pre_transform)
        self.model_name = model_name

        # Conditionally load the data
        if self.model_name:
            print(self.processed_paths[0])
            if not os.path.exists(self.processed_paths[0]):
                # Initialize the language model
                self.transformer = self.initialize_model()
                self.process()
            self.data, self.slices = torch.load(self.processed_paths[0])
        self.genome = self.parse_genome(genome)
        del genome

    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        # BUG we have to do this black magic because when you merge datasets with +
        # the genome is None
        if genome is None:
            return None
        else:
            data = {}
            data["gene_set"] = genome.gene_set
            return ParsedGenome(**data)

    def initialize_model(self):
        if self.model_name:
            return NucleotideTransformer()
        return None

    def process(self):
        if self.model_name is None:
            return

        data_list = []
        window_method, window_size, flag = self.MODEL_TO_WINDOW[self.model_name]

        # TODO check that genome gene set is SortedSet
        for gene_id in tqdm(self.genome.gene_set):
            sequence = self.genome[gene_id]

            if "utr" in window_method:
                dna_selection = getattr(sequence, window_method)(
                    window_size, allow_undersize=flag
                )
            else:
                dna_selection = getattr(sequence, window_method)(
                    window_size, is_max_size=flag
                )

            embeddings = self.transformer.embed(
                [dna_selection.seq], mean_embedding=True
            )

            # Create or update the dna_window dictionary
            dna_window_dict = {self.model_name: dna_selection}

            data = Data(id=gene_id, dna_windows=dna_window_dict)
            data.embeddings = {self.model_name: embeddings}
            data_list.append(data)

        if self.pre_transform:
            data_list = [self.pre_transform(data) for data in data_list]

        torch.save(self.collate(data_list), self.processed_paths[0])


if __name__ == "__main__":
    genome = SCerevisiaeGenome()
    model_names = [
        "nt_window_5979",
        # "nt_window_5979_max",
        # "nt_window_3utr_5979",
        # "nt_window_3utr_5979_undersize",
        # "nt_window_5utr_5979",
        # "nt_window_5utr_5979_undersize",
        # "nt_window_3utr_300",
        # "nt_window_3utr_300_undersize",
        # "nt_window_5utr_1000",
        # "nt_window_5utr_1000_undersize"
    ]
    datasets = []
    for model_name in model_names:
        dataset = NucleotideTransformerDataset(
            root="data/scerevisiae/nucleotide_transformer_embed",
            genome=genome,
            model_name=model_name,
        )
        datasets.append(dataset)
        print(f"Dataset for {model_name}: {dataset}")
    print(dataset)



import csv
import json
import os.path as osp

# Saccharomyces cerevisiae Raw Data Extraction Functions
from typing import Optional

import networkx as nx
import pandas as pd
import torch
from goatools.obo_parser import GODag
from torch_geometric.data import Data, Dataset, download_url
from torch_geometric.utils.convert import from_networkx

import torchcell.yeastmine as ym

# TODO write a data loader that can download the below files into ../data/
# This works already... just need to fully implement.
# download data from data_url and put in "data/Baryshnikova2010"
# import requests
# import os.path as osp

# url = "https://research.cs.umn.edu/csbio/SGAScore/Supplementary_data_1_SMF_standard_100209.xls"
# resp = requests.get(url)
# base_path = "data/Baryshnikova2010"
# file_name = url.split("/")[-1]
# with open(osp.join(base_path, file_name), "wb") as f:
#     f.write(resp.content)


def read_raw_data(
    singles: bool = True, digenic: bool = True, trigenic: bool = True
) -> dict:
    """[read raw data from data directory]

    Args:
        singles (bool, optional): [Include data]. Defaults to True.
        digenic (bool, optional): [Include data]. Defaults to True.
        trigenic (bool, optional): [Include Data]. Defaults to True.

    Returns:
        dict: [Keys are names of dataframes. Items are corresponding dataframe of raw fitness data.]
    """
    # initialize to None
    df_singles = None
    df_digenic_damp = None
    df_digenic_exe = None
    df_digenic_exn_nxe = None
    df_digenic_nxn = None
    df_trigenic = None

    df_names = []
    dfs = []
    # Read in Raw Data
    if singles:
        df_singles = pd.read_excel(
            r"data/Baryshnikova2010/10.1038-nmeth.1534_S1_Single-mutant_fitness_standard.xls",
            header=0,
            names=["gene", "fitness_mean", "fitness_std"],
        )
        df_names.append("df_singles")
        dfs.append(df_singles)
    if digenic:
        df_digenic_damp = pd.read_csv(
            r"data/TheCellMap/Data File S1. Raw genetic interaction datasets: Pair-wise interaction format/SGA_DAmP.txt",
            sep="\t",
        )
        df_names.append("df_digenic_damp")
        dfs.append(df_digenic_damp)
        df_digenic_exe = pd.read_csv(
            r"data/TheCellMap/Data File S1. Raw genetic interaction datasets: Pair-wise interaction format/SGA_ExE.txt",
            sep="\t",
        )
        df_names.append("df_digenic_exe")
        dfs.append(df_digenic_exe)
        df_digenic_exn_nxe = pd.read_csv(
            r"data/TheCellMap/Data File S1. Raw genetic interaction datasets: Pair-wise interaction format/SGA_ExN_NxE.txt",
            sep="\t",
        )
        df_names.append("df_digenic_exn_nxe")
        dfs.append(df_digenic_exn_nxe)
        df_digenic_nxn = pd.read_csv(
            r"data/TheCellMap/Data File S1. Raw genetic interaction datasets: Pair-wise interaction format/SGA_NxN.txt",
            sep="\t",
        )
        df_names.append("df_digenic_nxn")
        dfs.append(df_digenic_nxn)
        dfs.append(
            pd.concat(
                [df_digenic_damp, df_digenic_exe, df_digenic_exn_nxe, df_digenic_nxn]
            ).reset_index()
        )
        df_names.append("df_digenic_all")
    if trigenic:
        df_trigenic = pd.read_csv(
            r"data/Kuzmin2018/aao1729_Data_S1.tsv", sep="\t", encoding="utf-8"
        )
        df_names.append("df_trigenic")
        dfs.append(df_trigenic)

    raw_data = dict(
        zip(
            df_names,
            dfs,
        )
    )
    return raw_data


def get_gene_list(
    singles: bool = False, digenic: bool = True, trigenic: bool = True
) -> list:
    """[Get list of all gene names contained in dataframes. Intended to be able to get list of genes from all or subset of dataframes]

    Args:
        raw_data (dict): [Keys are names of dataframes. Items are data frames.]

    Returns:
        list: [Union of genes names from different dataframes]
    """
    gene_list_file = "data/preprocessed/gene_list.csv"
    if osp.exists(gene_list_file):
        with open("data/preprocessed/gene_list.csv", "r") as f:
            g_list = csv.reader(f, delimiter=",")
            gene_list = [row for row in g_list][0]
    else:
        raw_data = read_raw_data(singles, digenic, trigenic)
        data_name_list = []
        # Singles Data
        if "df_singles" in raw_data.keys():
            singles_names = (
                raw_data["df_singles"]["gene"].str.split("_", expand=True)[0].to_list()
            )
            data_name_list.append(pd.Series(singles_names))
        # Digenic Data
        if "df_digenic_damp" in raw_data.keys():
            digenic_damp_names = _ext_digenic_genes(raw_data["df_digenic_damp"])
            data_name_list.append(digenic_damp_names)
        if "df_digenic_exe" in raw_data.keys():
            digenic_exe_names = _ext_digenic_genes(raw_data["df_digenic_exe"])
            data_name_list.append(digenic_exe_names)
        if "df_digenic_exn_nxe" in raw_data.keys():
            digenc_exn_nxe_names = _ext_digenic_genes(raw_data["df_digenic_exn_nxe"])
            data_name_list.append(digenc_exn_nxe_names)
        if "df_digenic_nxn" in raw_data.keys():
            digenic_nxn_names = _ext_digenic_genes(raw_data["df_digenic_nxn"])
            data_name_list.append(digenic_nxn_names)
        # Trigenic Data
        if "df_trigenic" in raw_data.keys():
            trigenic_names = _ext_trigenic_genes(raw_data["df_trigenic"])
            data_name_list.append(trigenic_names)
        # Concatenate all data
        total_name_instances = pd.concat(data_name_list)
        gene_list = sorted(list(total_name_instances.drop_duplicates()))
        # first two genes not in SGD database, last three genes have stop codons in sequence, all have no, or few GO terms
        ignored_genes = get_ignored_genes()
        gene_list = [gene for gene in gene_list if gene not in ignored_genes]
        with open(gene_list_file, "w") as f:
            writer = csv.writer(f)
            writer.writerow(gene_list)
    return gene_list


def get_gene_id_translation_table():
    file_path = "data/preprocessed/gene_id_translation_table.csv"
    if osp.exists(file_path):
        df = pd.read_csv(file_path)
    else:
        print("NEED TO IMPLEMENT METHOD")
        df = None
    return df


def get_ignored_genes():
    ignored_genes = ["YIL080W", "YAR037W", "YAR040C", "YFL057C", "YFL056C"]
    return ignored_genes


def get_smf_bary(write_json=True, to_tensor=False) -> list:
    # TODO docstring
    data_path = "data/preprocessed/gene_singles_fitness_bary.json"
    if osp.exists(data_path):
        with open(data_path, "r") as f:
            gene_singles = json.load(f)
        if to_tensor:
            gene_singles = [val[0] for _, val in gene_singles]
            gene_singles = torch.tensor(list(gene_singles), dtype=torch.float)
        return gene_singles
    else:
        df = read_raw_data(True, False, False)["df_singles"]
        # only take full knockout alleles
        reg_allele_filter = [
            i == 1 for i in [len(i) for i in df["gene"].str.split("_")]
        ]
        df_reg_allele = df[reg_allele_filter]
        names = df_reg_allele["gene"].str.split("_", expand=True)[0].to_list()
        labels = df_reg_allele["fitness_mean"].to_list()
        label_data = [list(i) for i in zip(names, labels)]
        label_data = {i[0]: i[1] for i in label_data}
        # read gene_essentiality_raw.json to list of lists -> raw null mutant
        # TODO call lethality.whatever
        with open("data/preprocessed/gene_essentiality_raw.json", "r") as f:
            essential_genes = json.load(f)
        gene_singles = get_gene_list()
        gene_singles = {i: None for i in gene_singles}
        gene_list = get_gene_list()
        problem_label = {}
        for gene in gene_list:
            if essential_genes[gene] == "viable":
                try:
                    gene_singles[gene] = label_data[gene]
                except KeyError:
                    print("Warning! viable without fitness label")
                    problem_label[gene] = "viable"
                    # Cannot assign any value
                    gene_singles[gene] = -1
            elif essential_genes[gene] == "inviable":
                try:
                    gene_singles[gene] = label_data[gene]
                    problem_label[gene] = "inviable with fitness label"
                except KeyError:
                    gene_singles[gene] = 0
            elif essential_genes[gene] == "unknown":
                try:
                    gene_singles[gene] = label_data[gene]
                except KeyError:
                    print("Warning! unknown without fitness label")
                    problem_label[gene] = "unknown without fitness label"
                    # Cannot assign any value
                    gene_singles[gene] = -1
        # handling inviables
        problem_inviables = {
            k: v for k, v in problem_label.items() if v == "inviable with fitness label"
        }
        inviables_fitness = {
            gene: label_data[gene] for gene in problem_inviables.keys()
        }
        print(f"Fitness of null mutant for SGD 'inviable':\n {inviables_fitness}")
        gene_singles = [[[k], [v]] for k, v in gene_singles.items()]
        if write_json:
            # write gene_singles to json file in data/preprocessed
            with open(data_path, "w") as f:
                json.dump(gene_singles, f, indent=4)
        if to_tensor:
            gene_singles = [val[0] for _, val in gene_singles]
            gene_singles = torch.tensor(list(gene_singles), dtype=torch.float)
        return gene_singles


def get_double_fitness():
    pass


def get_triple_fitness():
    pass


# Hidden Functions
def _ext_digenic_genes(df_digenic: pd.DataFrame) -> pd.Series:
    """[Extract all digenic gene name instances from digenic interaction data]

    Args:
        df_digenic (pd.DataFrame): [digenic interaction data from data/TheCellMap]

    Returns:
        pd.Series: [(nx1) gene names with duplicates. Duplicates represent number of mutant instances]
    """
    qry = df_digenic["Query Strain ID"].str.split("_", expand=True)[0]
    arr = df_digenic["Array Strain ID"].str.split("_", expand=True)[0]
    names = pd.concat([qry, arr])
    return names


def _ext_trigenic_genes(df_trigenic: pd.DataFrame) -> pd.Series:
    """[Extract all trigenic gene name instances from trigenic interaction data]

    Args:
        df_trigenic (pd.DataFrame): [trigenic interaction data from data/Kuzmin2018]

    Returns:
        pd.Series: [(nx1) gene names with duplicates. Duplicates represent number of mutant instances]
    """
    qry_split = (
        df_trigenic["Query strain ID"]
        .str.split("_", expand=True)[0]
        .str.split("+", expand=True)
    )
    qry = pd.concat([qry_split[0], qry_split[1]])
    arr = df_trigenic["Array strain ID"].str.split("_", expand=True)[0]
    names = pd.concat([qry, arr])
    return names


if __name__ == "__main__":
    pass



# torchcell/losses/weighted_mse.py
# [[torchcell.losses.weighted_mse]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/losses/weighted_mse.py
# Test file: torchcell/losses/test_weighted_mse.py

import torch
import torch.nn as nn


class WeightedMSELoss(nn.Module):
    def __init__(self, mean_value: float, penalty: float = 1.0):
        super().__init__()
        self.mean_value = mean_value
        self.penalty = penalty

    def forward(self, y_pred, y_true):
        weights = torch.abs(y_true - self.mean_value)
        loss = torch.mean((self.penalty * 1 + weights) * (y_true - y_pred) ** 2)
        return loss


if __name__ == "__main__":
    mean_value = 0.85
    criterion = WeightedMSELoss(mean_value=mean_value)



# torchcell/datasets/protT5.py
# [[torchcell.datasets.protT5]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/protT5.py
# Test file: torchcell/datasets/test_protT5.py

import os
from collections.abc import Callable

import torch
from torch_geometric.data import Data
from tqdm import tqdm

from torchcell.datamodels import ModelStrictArbitrary
from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.models.protT5 import ProtT5
from torchcell.sequence import GeneSet, ParsedGenome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome


class ProtT5Dataset(BaseEmbeddingDataset):
    MODEL_TO_WINDOW = {
        "prot_t5_xl_uniref50_all": None,
        "prot_t5_xl_uniref50_no_dubious_uncharacterized": [
            "dubious",
            "uncharacterized",
        ],
        "prot_t5_xl_uniref50_no_dubious": ["Dubious"],
        "prot_t5_xl_uniref50_no_uncharacterized": ["Uncharacterized"],
    }

    def __init__(
        self,
        root: str,
        genome: SCerevisiaeGenome,
        model_name: str | None = None,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.genome = genome
        self.model_name = model_name
        super().__init__(root, self.model_name, transform, pre_transform)
        self.genome = self.parse_genome(genome)
        del genome

        # self.data, self.slices = torch.load(self.processed_paths[0])
        # self.data, self.slices = torch.load(
        #     self.processed_paths[0], map_location=self.device
        # )
        if self.model_name:
            if not os.path.exists(self.processed_paths[0]):
                self.transformer = self.initialize_transformer()
                self.process()
            # HACK we send cpu because all data needs to be on cpu for lightning
            # lightning automatically moves
            self.data, self.slices = torch.load(
                self.processed_paths[0], map_location="cpu"
            )

    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        if genome is None:
            return None
        else:
            data = {}
            data["gene_set"] = genome.gene_set
            return ParsedGenome(**data)

    def initialize_model(self) -> ProtT5:
        return ProtT5("prot_t5_xl_uniref50")

    def process(self):
        # HACK
        # self.transformer = self.initialize_model()
        if not self.model_name:
            return

        data_list = []

        exclude_classifications = self.MODEL_TO_WINDOW.get(self.model_name, None)
        for gene_id in tqdm(self.genome.gene_set):
            orf_classification = self.genome[gene_id].orf_classification[0]

            protein_sequence = str(self.genome[gene_id].protein.seq)

            if (
                exclude_classifications
                and orf_classification in exclude_classifications
            ):
                print(f"zeros for {gene_id}")
                embeddings = torch.zeros(1, 1024, dtype=torch.float32).to(self.device)
            else:
                embeddings = self.transformer.embed(
                    [protein_sequence], mean_embedding=True
                ).to(torch.float32)

            protein_data_dict = {self.model_name: protein_sequence}

            # Using 'dna_windows' for compatibility, but this might need a more general solution in the future
            data = Data(id=gene_id, dna_windows=protein_data_dict)
            data.embeddings = {self.model_name: embeddings}
            data_list.append(data)

        if self.pre_transform:
            data_list = [self.pre_transform(data) for data in data_list]

        torch.save(self.collate(data_list), self.processed_paths[0])


if __name__ == "__main__":
    import os.path as osp

    from dotenv import load_dotenv

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), overwrite=True
    )

    dataset = ProtT5Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/protT5_embed"),
        genome=genome,
        model_name="prot_t5_xl_uniref50_all",
    )

    dataset = ProtT5Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/protT5_embed"),
        genome=genome,
        model_name="prot_t5_xl_uniref50_no_dubious",
    )
